"""Compare VLDB settings of two objects. These objects can be a Dashboard,
DatasourceInstance, Document, IncrementalRefreshReport,Metric, OlapCube,
SuperCube, Project, Report or dictionary of object VLDB settings.
In this script settings were compared for Project, DatasourceInstance, OlapCube
and SuperCube. The function `compare_obj_vldbs` compares the VLDB settings of
two objects and returns keys were difference was spotted. If the `export_file`
parameter is provided, the differences will be exported to a csv file.

1. Connect to the environment.
2. List projects in the environment.
3. Select two projects for comparison.
4. List VLDB settings of the selected projects.
5. Compare VLDB settings of two projects.
6. List datasource instances.
7. Select two datasource instances for comparison.
8. Compare VLDB settings of two datasource instances.
9. List OLAP cubes.
10. Select two OLAP cubes for comparison.
11. Compare VLDB settings of two OLAP cubes.
12. List SuperCubes.
13. Select two SuperCubes for comparison.
14. Compare VLDB settings of two SuperCubes.
"""

import csv
from pprint import pprint

from mstrio.connection import get_connection
from mstrio.datasources import DatasourceInstance, list_datasource_instances
from mstrio.modeling import Metric
from mstrio.project_objects import (
    Dashboard,
    Document,
    IncrementalRefreshReport,
    Report,
)
from mstrio.project_objects.datasets import (
    OlapCube,
    SuperCube,
    list_olap_cubes,
    list_super_cubes,
)
from mstrio.server import Environment, Project


def get_vldb_settings_dict(obj) -> dict:
    """Get VLDB settings dictionary from the object."""
    if isinstance(obj, (Dashboard, Document, SuperCube, Report)):
        return obj.vldb_settings
    elif isinstance(
        obj,
        (DatasourceInstance, IncrementalRefreshReport, Metric, OlapCube, Project),
    ):
        return obj.list_vldb_settings(to_dictionary=True)
    else:
        return obj


def get_differing_part(dict1: dict, dict2: dict) -> dict:
    """Get differing part of two dictionaries."""
    return {k: dict1[k] for k in dict1 if dict1[k] != dict2.get(k)}


def compare_obj_vldbs(
    obj_1: (
        Dashboard
        | DatasourceInstance
        | Document
        | IncrementalRefreshReport
        | Metric
        | OlapCube
        | SuperCube
        | Project
        | Report
        | dict
    ),
    obj_2: (
        Dashboard
        | DatasourceInstance
        | Document
        | IncrementalRefreshReport
        | Metric
        | OlapCube
        | SuperCube
        | Project
        | Report
        | dict
    ),
    verbose: bool = False,
    comparison_level: int = 0,
    export_file: str = None,
) -> list[str]:
    """
    Compare VLDB settings of two objects.
    Args:
        obj_1(Dashboard, DatasourceInstance, Document, IncrementalRefreshReport,
            Metric, OlapCube, SuperCube, Project, Report, dict): First object to
            compare VLDB settings with.
        obj_2(Dashboard, DatasourceInstance, Document, IncrementalRefreshReport,
            Metric, OlapCube, SuperCube, Project, Report, dict): Second object
            to compare VLDB settings with.
        verbose(bool, optional): Controls the verbosity of the output. False by
            default.
        comparison_level(int, optional): Level comparison made inside the
            function. There are three levels:
                0 - for every setting compares only the value under the key
                    'value' and prints/exports if differs
                1 - compares all dict under the key and prints/exports only the
                    part that differs
                2 - compares all dict under the key and prints/exports all info
                    under that key if any difference is found
        export_file(str, optional): Path to a csv file where the differences
        will be exported if provided.

    Returns:
        List of keys with different settings.
    """

    if comparison_level not in [0, 1, 2]:
        comparison_level = 0

    obj_1_vldbs = get_vldb_settings_dict(obj_1)
    obj_2_vldbs = get_vldb_settings_dict(obj_2)

    if obj_1_vldbs.keys() != obj_2_vldbs.keys():
        if verbose:
            print(
                "Provided objects have different VLDB settings keys. Will "
                "compare only common keys."
            )
    else:
        if verbose:
            print("Provided objects have the same VLDB settings keys.")
    common_keys = obj_1_vldbs.keys() & obj_2_vldbs.keys()
    diff_keys = []

    for key in common_keys:
        if comparison_level == 0:
            if obj_1_vldbs[key].get('value') != obj_2_vldbs[key].get('value'):
                diff_keys.append(key)
        else:
            if obj_1_vldbs[key] != obj_2_vldbs[key]:
                diff_keys.append(key)
    if verbose:
        print("Keys with different values:")
        pprint(diff_keys)

    obj_1_header = (
        f'Object 1: {obj_1.name} | {obj_1.id}'
        if not isinstance(obj_1, dict)
        else 'Object 1'
    )

    obj_2_header = (
        f'Object 2: {obj_2.name} | {obj_2.id}'
        if not isinstance(obj_2, dict)
        else 'Object 2'
    )

    for key in diff_keys:
        if verbose:
            print(f"Difference for key '{key}':")
            if comparison_level == 0:
                print(obj_1_header)
                pprint(obj_1_vldbs[key]['value'])
                print(obj_2_header)
                pprint(obj_2_vldbs[key]['value'])
            elif comparison_level == 1:
                print(obj_1_header)
                pprint(get_differing_part(obj_1_vldbs[key], obj_2_vldbs[key]))
                print(obj_2_header)
                pprint(get_differing_part(obj_2_vldbs[key], obj_1_vldbs[key]))
            elif comparison_level == 2:
                print(obj_1_header)
                pprint(obj_1_vldbs[key])
                print(obj_2_header)
                pprint(obj_2_vldbs[key])
            print("\n")

    if export_file:
        with open(export_file, mode='w', newline='') as file:
            file_writer = csv.writer(
                file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL
            )
            file_writer.writerow(
                [
                    'Key',
                    obj_1_header,
                    obj_2_header,
                ]
            )
            for key in diff_keys:
                if comparison_level == 0:
                    file_writer.writerow(
                        [key, obj_1_vldbs[key]['value'], obj_2_vldbs[key]['value']]
                    )
                elif comparison_level == 1:
                    file_writer.writerow(
                        [
                            key,
                            get_differing_part(obj_1_vldbs[key], obj_2_vldbs[key]),
                            get_differing_part(obj_2_vldbs[key], obj_1_vldbs[key]),
                        ]
                    )
                elif comparison_level == 2:
                    file_writer.writerow([key, obj_1_vldbs[key], obj_2_vldbs[key]])
    return diff_keys


PROJECT_NAME = 'MicroStrategy Tutorial'  # Project to connect to
# Connect to environment without providing user credentials
# Variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Create an environment object
env = Environment(conn)
# List projects in the environment
projects = env.list_projects()

# Select two projects for comparison
project_1 = projects[0]
project_2 = projects[1]

# Compare VLDB settings of two projects
# Define export_file where exported data will be saved. In Workstation you need
# to provide absolute path to the file remembering about double backslashes on
# WIN for example: 'C:\\Users\\admin\\Documents\\data\\projects_vldbs_diff.csv'
FILE_PATH_PROJECT = '<PATH_TO_FILE>'
compare_obj_vldbs(
    project_1,
    project_2,
    verbose=True,
    comparison_level=1,
    export_file=FILE_PATH_PROJECT,
)

# List datasource instances
ds_instances = list_datasource_instances(connection=conn)

# Select two datasource instances for comparison
ds_1 = ds_instances[0]
ds_2 = ds_instances[1]

# Compare VLDB settings of two datasource instances
FILE_PATH_DATASOURCE = '<PATH_TO_FILE>'
compare_obj_vldbs(ds_1, ds_2, comparison_level=2, export_file=FILE_PATH_DATASOURCE)

# List OLAP cubes
olap_cubes = list_olap_cubes(connection=conn)

# Select two OLAP cubes for comparison
olap_cube_1 = olap_cubes[0]
olap_cube_2 = olap_cubes[1]

# Compare VLDB settings of two OLAP cubes
diff_keys = compare_obj_vldbs(olap_cube_1, olap_cube_2, comparison_level=1)
print(diff_keys)

# List SuperCubes
super_cubes = list_super_cubes(connection=conn)

# Select two SuperCubes for comparison
super_cube_1 = super_cubes[0]
super_cube_2 = super_cubes[1]
super_cube_1_vldbs = super_cube_1.vldb_settings

# Compare VLDB settings of two SuperCubes
compare_obj_vldbs(super_cube_1_vldbs, super_cube_2, verbose=True)



"""This is the demo script to show how administrator can manage items as they
are published to Library.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.dashboard import Dashboard
from mstrio.project_objects.library import Library


# Create connection based on workstation data
PROJECT_NAME = $project_name  # Project to connect to
conn = get_connection(workstationData, project_name=PROJECT_NAME)

DASHBOARD_ID = $dashboard_id  # Insert ID of Dashboard on which you want to perform actions

# Get single dashboard by its id
dashboard = Dashboard(connection=conn, id=DASHBOARD_ID)

# Publish dashboard to Library
dashboard.publish()

# List all items in Library as shortcuts
library = Library(connection=conn)
shortcuts = library.list_library_shortcuts()
for shortcut in shortcuts:
    print(f"Shortcut: {shortcut.name} (ID: {shortcut.id})")
    target = shortcut.target
    print(f"  Target Name: {target['name']}, Target ID: {target['id']}")

# List all dashboards in Library
dashboards_in_lib = library.dashboards
for dash in dashboards_in_lib:
    print(f"Dashboard in Library: {dash.name} (ID: {dash.id})")



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error getting Applications list.')
def get_applications(
    connection: Connection,
    output_flag: list[str] | None = None,
    fields: str | None = None,
) -> Response:
    """Get list of available Applications.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        output_flag (list, optional): Flag that specifies what should be
            included or filtered out of the application output
            Possible values:
                -DEFAULT
                -FILTER_AUTH_MODES
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    if output_flag is None:
        output_flag = ['DEFAULT']
    return connection.get(
        endpoint='/api/v2/applications',
        params={'outputFlag': output_flag, 'fields': fields},
    )


@ErrorHandler(err_msg='Error getting Application with ID: {id}.')
def get_application(
    connection: Connection,
    id: str,
    output_flag: list[str] | None = None,
    fields: str | None = None,
) -> Response:
    """Get an Application by its ID.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the application to get
        output_flag (list, optional): Flag that specifies what should be
            included or filtered out of the application output
            Possible values:
                -DEFAULT
                -INCLUDE_LOCALE
                -INCLUDE_ACL
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    if output_flag is None:
        output_flag = ['INCLUDE_LOCALE', 'INCLUDE_ACL']
    return connection.get(
        endpoint=f'/api/v2/applications/{id}',
        params={'outputFlag': output_flag, 'fields': fields},
    )


@ErrorHandler(err_msg='Error creating an Application.')
def create_application(
    connection: Connection,
    body: dict,
) -> Response:
    """Create a new Application.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new application

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.post(
        endpoint='/api/v2/applications',
        json=body,
    )


@ErrorHandler(err_msg='Error updating Application with ID: {id}.')
def update_application(
    connection: Connection,
    id: str,
    body: dict,
) -> Response:
    """Update an Application.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the application to update
        body (dict): JSON-formatted body of the updated application

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.put(
        endpoint=f'/api/v2/applications/{id}',
        json=body,
    )


@ErrorHandler(err_msg='Error deleting Application with ID: {id}.')
def delete_application(
    connection: Connection,
    id: str,
) -> Response:
    """Delete an Application.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the application to delete

    Returns:
        HTTP response object returned by the Strategy One REST server."""
    return connection.delete(
        endpoint=f'/api/v2/applications/{id}',
    )



import typing as t

from mstrio.api import projects as projects_api

if t.TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.server import Project


T = t.TypeVar('T')


# --- PARAMETERS Resolvers ---
def _get_id_from_params_set(
    object: T | str | None,
    object_id: str | None,
    object_prop: str | None,
    object_class: type[T],
    object_listing_method: t.Callable[[], list[T | dict]],
    fallback_value: str | None = None,
    property_name: str = 'name',
) -> str | None:
    """
    Utility function to get an object's ID from a set of parameters.

    Should be used inside a method that has a set of parameters that can hold
    the object's ID, but only one of those is required to be passed.

    Example:
        >>> # Method signature
        >>> def some_method(
        >>>     self,
        >>>     conn: Connection,
        >>>     project: Project | None = None,
        >>>     project_id: str | None = None,
        >>>     project_name: str | None = None,
        >>>     ...
        >>> ): ...
        >>>
        >>> # Method body
        >>> project_id = get_id_from_params_set(
        >>>     project,
        >>>     project_id,
        >>>     project_name,
        >>>     Project,
        >>>     lambda: list_projects(conn),
        >>>     fallback_value=conn.project_id,
        >>> )
        >>> ...

    Args:
        object: An instance of the `object_class` or its ID or name.
        object_id: ID of the object.
        object_prop: `property_name` of the object.
        object_class: The class of the object.
        object_listing_method: A method that returns a list of all objects of
            the given type.
        fallback_value: A value to return if none of the parameters are set.
        property_name: The name of the property to use for matching.

    Returns:
        The ID of the object if found, otherwise the `fallback_value`.
    """

    if object is not None:
        if isinstance(object, object_class):
            return object.id
        return _get_id_from_params_set(
            object=None,
            object_id=object,  # assume ID was provided as `object`
            object_prop=None,
            object_class=object_class,
            object_listing_method=object_listing_method,
            fallback_value=fallback_value,
            property_name=property_name,
        )
    if object_id is not None:
        from mstrio.utils.helper import is_valid_str_id

        if is_valid_str_id(object_id):
            return object_id
        return _get_id_from_params_set(
            object=None,
            object_id=None,
            object_prop=object_id,  # assume `object_prop` was provided as `object_id`
            object_class=object_class,
            object_listing_method=object_listing_method,
            fallback_value=fallback_value,
            property_name=property_name,
        )
    if object_prop is not None:
        items = object_listing_method()

        assert isinstance(items, (list, tuple)), (
            f"Expected listing method to return a list or tuple, "
            f"but found: {type(items).__name__}"
        )
        assert all(isinstance(obj, (object_class, dict)) for obj in items), (
            f"Expected all objects to be of type {object_class.__name__} or pure dicts,"
            f" but found types are: {{type(obj).__name__ for obj in items}}"
        )

        def get_prop_val(itm):
            if isinstance(itm, object_class):
                return getattr(itm, property_name, None)
            return itm.get(property_name)

        valid_objects = [obj for obj in items if get_prop_val(obj) == object_prop]

        assert len(valid_objects) == 1, (
            f"Could not uniquely identify the {object_class.__name__} "
            f"by it's {property_name} '{object_prop}'. "
            f"Found {len(valid_objects)} items. "
            f"Please provide a {object_class.__name__} class instance or ID instead."
        )

        itm = valid_objects[0]
        return itm.id if isinstance(itm, object_class) else itm.get('id')

    return fallback_value


def get_project_id_from_params_set(
    conn: 'Connection | None',
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    assert_id_exists: bool = True,
    no_fallback_from_connection: bool = False,
) -> str | None:
    """
    Utility function to get a project's ID from a set of parameters.

    Should be used inside a method that has a set of parameters that can hold
    the project's ID, but only one of those is required to be passed.

    Args:
        conn: A Strategy connection object.
        project: An instance of the `Project` class or its ID or name.
        project_id: ID of the project.
        project_name: Name of the project.
        assert_id_exists: If `False`, return `None` if no project is found.
            If `True`, raises an `AssertionError` if no project is found.
        no_fallback_from_connection: If `True`, the function will return `None`
            instead of the connection's default project ID if no project is
            found.

    Returns:
        The ID of the project if found, otherwise the connection's default
        project ID, unless `no_fallback_from_connection` is True.
    """

    from mstrio.server import Project

    try:
        ret = _get_id_from_params_set(
            project,
            project_id,
            project_name,
            Project,
            lambda: projects_api.get_projects(conn).json() if conn else [],
            fallback_value=(
                conn.project_id if not no_fallback_from_connection and conn else None
            ),
        )
    except AssertionError as err:
        raise ValueError(err) from err

    if assert_id_exists and not ret:
        from mstrio.utils.helper import exception_handler

        msg = (
            "Could not determine the project ID. Please provide a Project "
            "class instance or ID directly."
        )
        exception_handler(msg, exception_type=ValueError)

    return ret


# FYI: Copy-pastable parts for project-related parameters:
"""
[params to the method]
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,

[DOCSTRING for "Args" part]
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name

[implementation of usual use]
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
        )
"""

# TODO: add params `project_id` and `project_name` to where
# we have only `project`

# --- END: PARAMETERS Resolvers ---


# --- FILTERS KWARGS Resolvers ---
def validate_owner_key_in_filters(filters_kwargs: dict[str, t.Any]) -> None:
    """Validate if `owner` filter kwarg is in valid shape.

    Note:
        REST requires the shape to be `{'id': <id>}`. However, not everywhere!
        There are places where `owner` needs to be an ID and places where it
        needs to be string of exact match on name. Use at your own discretion.

    Note:
        This method does not return a value, it modifies the input dict
        in-place!

    Args:
        filters_kwargs: Dictionary of filter kwargs to validate `owner` in.
    """
    if 'owner' not in filters_kwargs:
        return  # NOOP

    val = filters_kwargs['owner']
    if val is None:
        del filters_kwargs['owner']
        return

    from mstrio.utils.entity import Entity

    if isinstance(val, Entity):
        # We do not care for more specific module than `Entity`
        # REST will handle invalid ID
        val = val.id

    if isinstance(val, str):
        filters_kwargs['owner'] = {"id": val}
        return

    if isinstance(val, dict) and "id" in val:
        filters_kwargs['owner'] = {"id": val["id"]}  # sanitize: remove other keys
        return

    raise AttributeError(
        "`owner` filter has incorrect value. It should be a class instance, "
        "item ID as string or dict in a shape `{'id': <id>}`"
    )


# --- END: FILTERS KWARGS Resolvers ---



from typing import TYPE_CHECKING

from mstrio.api import subscriptions
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.version_helper import method_version_handler

if TYPE_CHECKING:
    from mstrio.distribution_services import Event, Schedule, Subscription  # noqa: F401
    from mstrio.project_objects import Dashboard, Report  # noqa: F401
    from mstrio.users_and_groups import User  # noqa: F401

RelatedSubscriptionTypes = 'User | Dashboard | Report | Schedule | Event'


class RelatedSubscriptionMixin:
    """RelatedSubscriptionMixin class adds listing support for supported
    objects."""

    @method_version_handler('11.4.0600')
    def list_related_subscriptions(
        self: RelatedSubscriptionTypes, to_dictionary: bool = False
    ) -> list['Subscription'] | list[dict]:
        """List all subscriptions that are dependent on the object.

        Args:
            to_dictionary (bool, optional): If True returns a list of
                subscription dicts, otherwise (default) returns a list of
                subscription objects
        """
        object_type = self.__class__.__name__.lower()

        if object_type in ['user', 'event', 'schedule']:
            project_id = None
        else:
            project_id = get_project_id_from_params_set(
                self.connection,
                self.project_id,
                None,
                None,
            )

        objects = (
            subscriptions.get_dependent_subscriptions(
                self.connection, self.id, object_type, project_id
            )
            .json()
            .get('subscriptions', [])
        )

        if to_dictionary:
            return objects

        from mstrio.distribution_services.subscription.subscription_manager import (
            dispatch_from_dict,
        )

        return [
            dispatch_from_dict(
                source=obj,
                connection=self.connection,
                project_id=project_id or obj.get('project', {}).get('id'),
            )
            for obj in objects
        ]



"""
Manage project settings.
Change project settings, copy settings from one project to another.
Export project settings to CSV.

1. Connect to the environment using data from the workstation
2. Get project object based on provided id
3. Get another project object based on provided id
4. Change single setting for a project
5. Apply all settings from one project to another
6. Export settings to a CSV file and import them to another project
7. Export settings to a CSV file with their descriptions
"""

from mstrio.connection import get_connection
from mstrio.server import Project

conn = get_connection(workstationData, 'MicroStrategy Tutorial')

project = Project(connection=conn, id='B3FEE61A11E696C8BD0F0080EFC58F44')
test_project = Project(connection=conn, id='B7CA92F04B9FAE8D941C3E9B7E0CD754')

# Change single setting for a project
# You can get a list of all available settings by calling:
# project.settings.list_properties()
project.settings.maxMstrFileSize = 50
project.settings.update()

# Apply all settings from one project to another
project.settings.alter(**test_project.settings.list_properties(show_names=False))
project.settings.update()

# Or export settings to a CSV file and import them to another project
settings_file_name = 'settings.csv'
test_project.settings.to_csv(settings_file_name)
project.settings.import_from(settings_file_name)
project.settings.update()

# Export settings to a CSV file with their descriptions
project.settings.to_csv('settings_file_name', show_description=True)



import inspect
import logging
import os
from base64 import b64encode
from datetime import datetime
from enum import IntEnum
from getpass import getpass
from typing import TYPE_CHECKING

import requests
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from requests.cookies import RequestsCookieJar

from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.resolvers import get_project_id_from_params_set

if TYPE_CHECKING:
    from urllib3 import disable_warnings
    from urllib3.exceptions import InsecureRequestWarning

    from mstrio.project_objects.applications import Application
    from mstrio.server.project import Project
else:
    from requests.packages.urllib3.exceptions import InsecureRequestWarning
    from requests.packages.urllib3 import disable_warnings

from mstrio import config
from mstrio.api import authentication, hooks, misc
from mstrio.helpers import IServerError, IServerException, VersionException
from mstrio.utils import helper, sessions

logger = logging.getLogger(__name__)

APPLICATION_TYPE_PYTHON = 76
APPLICATION_TYPE_WEB = 35

try:
    # assigned right away instead of being calculated in a method so that
    # if a user creates their own `workstationData` variable, this will still
    # show false as this is run before the script is executed in most cases
    _is_in_workstation: bool | None = any(
        'workstationData' in dict(inspect.getmembers(frame[0]))['f_globals']
        for frame in inspect.stack()
    )
except Exception:
    # This is not essential enough info to try to handle it
    # in a more sophisticated way
    _is_in_workstation = None


class LoginMode(IntEnum):
    STANDARD = 1
    ANONYMOUS = 8
    LDAP = 16
    API_TOKEN = 4096


def get_connection(
    workstation_data: dict,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    ssl_verify: bool = False,
) -> 'Connection | None':
    """Connect to environment without providing user's credentials.

    It is possible to provide `project`, `project_id` or `project_name` to
    select project. When all are `None`, project selection is cleared.
    Project can be also selected later by calling method
    `select_project` on Connection object.

    Note:
        `ssl_verify` is set to False by default just for the `get_connection`
        function as it is designed for usage inside Workstation.
        When `ssl_verify` is set to False, warning about missing certificate
        verification (InsecureRequestWarning) is disabled.

    Args:
        workstation_data (object): object which is stored in a 'workstationData'
            variable within Workstation
        project (Project, str, optional): Project object or ID or name
            of the project to be selected
        project_id (str, optional): ID of project to be selected
        project_name (str, optional): Name of project to be selected
        ssl_verify (bool, optional): If False (default), does not verify the
            server's SSL certificates

    Returns:
        connection to I-Server or None in case of some error
    """
    if not ssl_verify:
        disable_warnings(category=InsecureRequestWarning)

    try:
        logger.info('Creating connection from Workstation Data object...')
        # get base url from Workstation Data object
        base_url = helper.url_check(workstation_data['defaultEnvironment']['url'])
        # get headers from Workstation Data object
        headers = workstation_data['defaultEnvironment']['headers']
        # get cookies from Workstation Data object
        cookies = workstation_data['defaultEnvironment']['cookies']

        # prepare cookies for request
        jar = RequestsCookieJar()
        for c in cookies:
            cookie_values = {'domain': '', 'path': '/'}
            cookie_values.update(c)
            jar.set(
                cookie_values['name'],
                cookie_values['value'],
                domain=cookie_values['domain'],
                path=cookie_values['path'],
            )
    except Exception as e:
        logger.error(
            f'Some error occurred while preparing data to get identity token: \n{e}'
        )
        return None

    # get identity token
    response = requests.post(
        base_url + '/api/auth/identityToken',
        headers=headers,
        cookies=jar,
        verify=ssl_verify,
    )
    if response.ok:
        # create connection to I-Server
        conn = Connection(
            base_url,
            identity_token=response.headers['X-MSTR-IdentityToken'],
            project=project,
            project_id=project_id,
            project_name=project_name,
            ssl_verify=ssl_verify,
        )
        conn._through_get_connection = True
        return conn

    logger.error(
        f'HTTP {response.status_code} - {response.reason}, Message {response.text}'
    )

    return None


class Connection:
    """Connect to and interact with the Strategy One environment.

    Creates a connection object which is used in subsequent requests and
    manages the user's connection with the Strategy One REST and Intelligence
    Servers.
    The connection is automatically renewed, or reconnected if server's session
    associated with the connection expires due to inactivity.

    Examples:
        >>> from mstrio import connection
        >>>
        >>> # connect to the environment and chosen project
        >>> conn = connection.Connection(
        >>>     base_url="https://demo.microstrategy.com/MicroStrategyLibrary",
        >>>     username="username",
        >>>     password="password",
        >>>     project_name="MicroStrategy Tutorial"
        >>> )
        >>> # disconnect
        >>> conn.close()

    Attributes:
        base_url: URL of the Strategy One REST API server.
        username: Username.
        project_id: Id of the connected Strategy One Project.
        project_name: Name of the connected Strategy One Project.
        login_mode: Authentication mode. Standard = 1 (default), LDAP = 16
            or API Token = 4096.
        ssl_verify: If True (default), verifies the server's SSL certificates
            with each request.
        user_id: Id of the authenticated user
        user_full_name: Full name of the authenticated user
        user_initials: Initials of the authenticated user
        iserver_version: Version of the I-Server
        web_version: Version of the Web Server
        token: authentication token
        working_set: Number of report/document instances that are kept in memory
            on the server before the oldest instance is replaced.
        max_search: Maximum number of concurrent searches.
        timeout: time after the server's session expires, in seconds
    """

    def __init__(
        self,
        base_url: str,
        username: str | None = None,
        password: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        login_mode: int | LoginMode | None = None,
        ssl_verify: bool = True,
        certificate_path: str | None = None,
        proxies: dict | None = None,
        identity_token: str | None = None,
        api_token: str | None = None,
        application_id: 'str | Application | None' = None,
        working_set: int | None = None,
        max_search: int | None = None,
        verbose: bool = True,
    ):
        """Establish a connection with Strategy One REST API.

        You can establish connection by either providing set of values
        (`username`, `password`, `login_mode`), just `identity_token`
        or just `api_token`.

        When `api_token` is provided and `login_mode` is set to 4096
        or omitted, the connection is established using API Token and
        all other authentication parameters are ignored.

        Note:
            When project cannot be established or is not provided, it is reset
            to `None`.

        Args:
            base_url (str): URL of the Strategy One REST API server.
                Typically of the form:
                "https://<mstr_env>.com/MicroStrategyLibrary/api"
            username (str, optional): Username
            password (str, optional): Password
            project_name (str, optional): Name of the project you intend
                to connect to (case-sensitive). Provide either Project ID
                or Project Name.
            project_id (str, optional): Id of the project you intend to
                connect to (case-sensitive). Provide either Project ID
                or Project Name.
            login_mode (int | LoginMode, optional): Specifies the
                authentication mode to use. Supported authentication
                modes are: Standard (1) (default), LDAP (16), or
                API Token (4096). Defaults to 1 unless `api_token` is
                provided, then 4096.
            ssl_verify (bool, optional): If True (default), verifies the
                server's SSL certificates with each request
            certificate_path (str, optional): Path to SSL certificate file, if
                None and ssl_verify is True then the certificate will be looked
                for in current working directory
            proxies (dict, optional): Dictionary mapping protocol or protocol
                and host to the URL of the proxy (e.g. {'http': 'foo.bar:3128',
                'http://host.name': 'foo.bar:4012'})
            identity_token (str, optional): Identity token for delegated
                session. Used for connection initialized by GUI.
            api_token (str, optional): API token for authenticating using
                `login_mode` 4096. For other login modes it is ignored.
            application_id (str, Application, optional): Login using the
                login mode configured in a custom application. If it's
                not provided, the server login modes are used.
            working_set (int, optional): Number of report/document instances
                that are kept in memory on the server before the oldest instance
                is replaced. Minimum value is 3, if None, will be set to 10 by
                default.
            max_search (int, optional): Maximum number of concurrent searches.
                If None, will be set to 3 by default.
            verbose (bool, optional): True by default. Controls the amount of
                feedback from the I-Server.
        """

        if login_mode is None:
            login_mode = 4096 if api_token else 1

        login_mode = get_enum_val(login_mode, LoginMode)

        helper.validate_param_value(
            'working_set', working_set, int, min_val=3, special_values=[None]
        )
        if max_search is not None:
            helper.validate_param_value('max_search', max_search, int)

        # set the verbosity globally
        config.verbose = bool(verbose and config.verbose)

        self.base_url = helper.url_check(base_url)
        self.username = username
        self.login_mode = login_mode
        self.certificate_path = certificate_path
        self.identity_token = identity_token
        self.api_token = api_token
        self._session = self.__configure_session(ssl_verify, certificate_path, proxies)
        self._web_version = None
        self._iserver_version = None
        self._user_id = None
        self._user_full_name = None
        self._user_initials = None
        self.__password = password
        self.last_active = None
        self.timeout = None
        self.working_set = working_set
        self.max_search = max_search
        self.__through_get_connection: bool = False

        # do not check application type if delegated
        self._application_type = (
            APPLICATION_TYPE_PYTHON if identity_token is None else None
        )

        from mstrio.project_objects.applications import Application

        self.application_id = (
            application_id.id
            if isinstance(application_id, Application)
            else application_id
        )

        # delegate identity token or connect and create new session
        if self.identity_token:
            self.delegate()
        else:
            self.connect()

        if not self.__check_version():
            msg = (
                f'This version of mstrio is only supported on MicroStrategy 11.1.0400 '
                f'or higher.\n'
                f'Current Intelligence Server version: {self._iserver_version}\n'
                f'Current MicroStrategy Web version: {self._web_version}'
            )
            logger.warning(msg)
            helper.exception_handler(
                msg='MicroStrategy Version not supported.',
                exception_type=VersionException,
            )
            return

        from mstrio.utils.version_helper import is_server_min_version

        # In some versions APPLICATION_TYPE_PYTHON is supported at login,
        # but several features are not available. Reconnect with
        # APPLICATION_TYPE_WEB if the server version is not compatible
        if (
            self._application_type
            and self._application_type != APPLICATION_TYPE_WEB
            and not is_server_min_version(self, '11.5.0600')
        ):
            self._application_type = APPLICATION_TYPE_WEB
            self.token = None
            self.connect()

        if application_id and not is_server_min_version(self, '11.3.1200'):
            logger.warning(
                "The `application_id` argument requires iServer version 11.3.1200 "
                f"or later. Since your server version: {self._iserver_version} "
                "is not compatible, this parameter will be omitted."
            )
            self.application_id = None

        self.project_id = None
        self.select_project(project, project_id, project_name)

    @property
    def _through_get_connection(self) -> bool:
        """Indicates if the connection was established via `get_connection`."""
        return self.__through_get_connection

    @_through_get_connection.setter
    def _through_get_connection(self, value: bool) -> None:
        if value is True and not self.is_run_in_workstation():
            raise SyntaxError(
                "Connection established via `get_connection` can only be used "
                "inside Workstation."
            )
        self.__through_get_connection = value

    def __enter__(self):
        if self._through_get_connection:
            raise SyntaxError(
                "Connection established via `get_connection` cannot be used "
                "as a context manager (aka. using `with` syntax) as running "
                "the script would disconnect you from Workstation."
            )
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def connect(self) -> None:
        """Authenticates the user and creates a new connection with the
        Intelligence Server.

        If an active connection is detected, the session is renewed.
        """
        response = self._renew() if self.token else None

        if response and response.ok:
            self._reset_timeout()
            if config.verbose:
                logger.info(
                    'Connection to Strategy One Intelligence Server was renewed.'
                )
        else:
            response = self._login()
            self._reset_timeout()
            self.token = response.headers['X-MSTR-AuthToken']
            self._fetch_session_info()

            if config.verbose:
                logger.info(
                    'Connection to Strategy One Intelligence Server has been '
                    'established.'
                )

    renew = connect

    def delegate(self):
        """Delegates identity token to get authentication token and connect to
        Strategy One Intelligence Server."""
        response = authentication.delegate(
            self, self.identity_token, whitelist=[('ERR003', 401)]
        )
        if response.ok:
            self._reset_timeout()
            self.token = response.headers['X-MSTR-AuthToken']
            self._fetch_session_info()
            if config.verbose:
                logger.info(
                    'Connection with Strategy One Intelligence Server has been '
                    'delegated.'
                )
        else:
            logger.warning(
                "Could not share existing connection session, please input credentials:"
            )
            self.__prompt_credentials()
            self.connect()

    def get_identity_token(self) -> str:
        """Create new identity token using existing authentication token."""
        response = authentication.identity_token(self)
        return response.headers['X-MSTR-IdentityToken']

    def get_api_token(self) -> str:
        """Create new API token using existing authentication token."""
        response = authentication.api_token(self)
        return response.json()['apiToken']

    def validate_identity_token(self) -> bool:
        """Validate the identity token."""
        validate = authentication.validate_identity_token(self, self.identity_token)
        return validate.ok

    def close(self):
        """Closes a connection with Strategy One REST API."""
        if self._through_get_connection:
            raise SyntaxError(
                "Connection established via `get_connection` cannot be closed "
                "as it would disconnect you from Workstation."
            )

        authentication.logout(connection=self, whitelist=[('ERR009', 401)])

        self._session.close()

        self.token = None

        if config.verbose:
            logger.info(
                'Connection to Strategy One Intelligence Server has been closed.'
            )

    def status(self) -> bool:
        """Checks if the session is still alive.

        Raises:
            HTTPError if I-Server behaves unexpectedly
        """
        status = self._status()

        if status.status_code == 200:
            logger.info('Connection to Strategy One Intelligence Server is active.')
            return True
        else:
            logger.info('Connection to Strategy One Intelligence Server is not active.')
            return False

    def select_project(
        self,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> None:
        """Select project for the given connection based.

        Args:
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Raises:
            ValueError: if project with given id or name does not exist or
                cannot find a unique project with given name
        """

        proj_id = get_project_id_from_params_set(
            self,
            project,
            project_id,
            project_name,
            assert_id_exists=False,
            no_fallback_from_connection=True,
        )

        if not proj_id:
            self.project_id = None
            self._session.headers['X-MSTR-ProjectID'] = None

            if config.verbose:
                logger.info('No Project selected in Connection object.')

            return None

        if config.verbose:
            from mstrio.server.project import Project

            logger.info('Project selected in Connection object:')
            Project(self, id=proj_id)  # this will log the project info

        self.project_id = proj_id
        self._session.headers['X-MSTR-ProjectID'] = self.project_id

    @sessions.log_request(logger)
    @sessions.renew_session
    def _request(
        self, method_name: str, url: str, endpoint: str, **kwargs
    ) -> requests.Response:
        if url and endpoint:
            raise ValueError(
                'Either `url` or `endpoint` argument should be provided, not both.'
            )

        if endpoint:
            url = self.base_url + endpoint

        name2method = {
            'GET': self._session.get,
            'POST': self._session.post,
            'PUT': self._session.put,
            'PATCH': self._session.patch,
            'DELETE': self._session.delete,
            'HEAD': self._session.head,
        }
        method = name2method[method_name.upper()]

        return method(url, **kwargs)

    def get(self, url=None, *, endpoint=None, **kwargs):
        """Sends a GET request."""
        return self._request('GET', url, endpoint, **kwargs)

    def post(self, url=None, *, endpoint=None, **kwargs):
        """Sends a POST request."""
        return self._request('POST', url, endpoint, **kwargs)

    def put(self, url=None, *, endpoint=None, **kwargs):
        """Sends a PUT request."""
        return self._request('PUT', url, endpoint, **kwargs)

    def patch(self, url=None, *, endpoint=None, **kwargs):
        """Sends a PATCH request."""
        return self._request('PATCH', url, endpoint, **kwargs)

    def delete(self, url=None, *, endpoint=None, **kwargs):
        """Sends a DELETE request."""
        return self._request('DELETE', url, endpoint, **kwargs)

    def head(self, url=None, *, endpoint=None, **kwargs):
        """Sends a HEAD request."""
        return self._request('HEAD', url, endpoint, **kwargs)

    def _status(self):
        return authentication.session_status(connection=self)

    def _login(self):
        try:
            return authentication.login(connection=self)
        except IServerError as e:
            if 'constraint violations' in str(e):
                # The error indicates APPLICATION_TYPE_PYTHON is not supported
                logger.error(
                    'Server does not support new Python Application Type '
                    f'({APPLICATION_TYPE_PYTHON}). Retrying connection with old '
                    f'setup: WEB Application Type ({APPLICATION_TYPE_WEB}).'
                )

                self._application_type = APPLICATION_TYPE_WEB
                return authentication.login(connection=self)
            else:
                raise e

    def _renew(self):
        return authentication.session_renew(connection=self)

    def _renew_or_reconnect(self):
        response = self._renew() if self.token else None

        if not self.identity_token and (not response or not response.ok):
            response = self._login()
            self.token = response.headers['X-MSTR-AuthToken']

        if response and response.ok:
            self._reset_timeout()

    def _fetch_session_info(self):
        res = self._status()
        json_data = res.json() if res.ok else {}
        self.timeout = json_data.get('timeout')
        self.working_set = json_data.get('workingSet')
        self.max_search = json_data.get('maxSearch')

    def _reset_timeout(self):
        self.last_active = datetime.now()

    def _is_session_expired(self):
        return (datetime.now() - self.last_active).seconds > self.timeout

    def _get_authorization(self) -> str:
        self.__prompt_credentials()
        credentials = f"{self.username}:{self.__password}".encode()
        encoded_credential = b64encode(credentials)
        auth = "Basic " + str(encoded_credential, 'utf-8')
        return auth

    def _validate_project_selected(self) -> None:
        if self.project_id is None:
            raise AttributeError(
                "Project not selected. Select project using `select_project` method."
            )

    def __prompt_credentials(self) -> None:
        self.username = self.username or input("Username: ")
        self.__password = self.__password or getpass("Password: ")
        self.login_mode = self.login_mode or int(
            input("Login mode (1 - Standard, 16 - LDAP): ")
        )

    def __check_version(self) -> bool:
        """Checks version of I-Server and Strategy One Web and store these
        variables."""

        resp = misc.server_status(self)
        if not resp:
            logger.warning(
                "Could not read status of I-Server. "
                "Assuming misconfiguration of Library and "
                "Server Version of 11.1.0400 or higher. "
                "Some functionalities may not work properly. "
                "Please check configuration of Library and I-Server availability."
            )
            return True

        err_msg = (
            "I-Server is currently unavailable or connection creation failed. "
            "Please verify or contact your administrator."
        )

        if not resp.ok:
            raise IServerException(err_msg)

        json_response = resp.json()
        if (
            not json_response
            or "iServerVersion" not in json_response
            or "webVersion" not in json_response
        ):
            raise IServerException(err_msg)

        self._iserver_version = json_response["iServerVersion"][:9]
        self._web_version = json_response["webVersion"][:9]

        from mstrio.utils.version_helper import meets_minimal_version

        ret = meets_minimal_version(self._iserver_version, "11.1.0400")

        if ret and self._iserver_version:
            # save the version of IServer in config file
            config.iserver_version = self._iserver_version

        return ret

    def __configure_session(
        self,
        verify,
        certificate_path,
        proxies,
        existing_session=None,
        retries=2,
        backoff_factor=0.3,
    ):
        """Creates a shared requests.Session() object with configuration from
        the initialization. Additional parameters change how the HTTPAdapter is
        configured.

        Args:
            existing_session (requests.Session() object, optional): Optional
                existing Session object to configure
            retries (int, optional): number of retries for certain type of
                requests (i.e. get)
            backoff_factor (float, optional):  backoff factor to apply between
                attempts after the second try (most errors are resolved
                immediately by a second try without a delay). urllib3 will sleep
                for: {backoff factor} * (2 ^ ({number of total retries} - 1))
                seconds. If the backoff_factor is 0.1, then sleep() will sleep
                for [0.0s, 0.2s, 0.4s, ...] between retries. It will never be
                longer than Retry. BACKOFF_MAX. By default, backoff is disabled
                (set to 0).
        """
        session = existing_session or Session()
        session.proxies = proxies or {}
        session.verify = self._configure_ssl(verify, certificate_path)

        response_hooks = []
        if config.debug:
            response_hooks.append(hooks.print_url)
        if config.save_responses:
            response_hooks.append(hooks.save_response)
        if response_hooks:
            session.hooks['response'] = response_hooks

        status_forcelist = frozenset([413, 429, 500, 503])  # retry after status

        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

        return session

    @staticmethod
    def _configure_ssl(ssl_verify, certificate_path):
        def get_certs_from_cwd():
            cert_extensions = ['crt', 'pem', 'p12']
            return [
                file
                for file in os.listdir('.')
                if file.split('.')[-1] in cert_extensions
            ]

        def find_cert_in_cwd(ssl_verify):
            certs = get_certs_from_cwd()
            if certs and ssl_verify:
                return certs[0]
            else:
                return ssl_verify

        if certificate_path:
            return certificate_path
        return find_cert_in_cwd(ssl_verify)

    def __get_user_info(self) -> None:
        response = authentication.get_info_for_authenticated_user(
            connection=self
        ).json()
        self._user_id = response.get("id")
        self._user_full_name = response.get("fullName")
        self._user_initials = response.get("initials")

    @classmethod
    def is_run_in_workstation(cls) -> bool | None:
        """Check if the script is run in Workstation.

        Returns:
            bool: True if the script is run in Workstation, False otherwise.
                It may return None when gathering this info failed.
        """
        return _is_in_workstation

    @property
    def project_name(self) -> str | None:
        if self.project_id is None:
            return None

        from mstrio.server.project import Project

        return Project(self, id=self.project_id).name

    @property
    def user_id(self) -> str:
        if not self._user_id:
            self.__get_user_info()
        return self._user_id

    @property
    def user_full_name(self) -> str:
        if not self._user_full_name:
            self.__get_user_info()
        return self._user_full_name

    @property
    def user_initials(self) -> str:
        if not self._user_initials:
            self.__get_user_info()
        return self._user_initials

    @property
    def web_version(self) -> str | None:
        if not self._web_version:
            self.__check_version()

        return self._web_version

    @property
    def iserver_version(self) -> str | None:
        if not self._iserver_version:
            self.__check_version()

        return self._iserver_version

    @property
    def token(self) -> str:
        return self._session.headers.get('X-MSTR-AuthToken')

    @token.setter
    def token(self, token: str) -> str:
        self._session.headers['X-MSTR-AuthToken'] = token
        return token

    @property
    def headers(self):
        return self._session.headers

    @property
    def environment(self):
        from mstrio.server.environment import Environment

        return Environment(connection=self)



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(err_msg="Error listing Transmitters.")
def get_transmitters(connection: 'Connection', error_msg: str | None = None):
    """Get a list of all transmitters that the authenticated user has access
    to.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    return connection.get(endpoint='/api/transmitters')


@ErrorHandler(err_msg="Error creating Transmitter.")
def create_transmitter(connection: 'Connection', body, error_msg: str | None = None):
    """Create a new transmitter.

    Args:
        connection: Strategy One REST API connection object
        body: Transmitter creation body
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    return connection.post(endpoint='/api/transmitters', json=body)


@ErrorHandler(err_msg="Error getting Transmitter with ID {id}")
def get_transmitter(connection: 'Connection', id: str, error_msg: str | None = None):
    """Get transmitter by a specific id.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the transmitter
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    return connection.get(endpoint=f'/api/transmitters/{id}')


@ErrorHandler(err_msg="Error updating Transmitter with ID {id}")
def update_transmitter(
    connection: 'Connection', id: str, body: dict, error_msg: str | None = None
):
    """Update a transmitter.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the transmitter
        body: Transmitter update info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    return connection.put(endpoint=f'/api/transmitters/{id}', json=body)


@ErrorHandler(err_msg="Error deleting Transmitter with ID {id}")
def delete_transmitter(connection: 'Connection', id: str, error_msg: str | None = None):
    """Delete a transmitter.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the transmitter
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    return connection.delete(endpoint=f'/api/transmitters/{id}')



"""Delete all subscription in all projects which owners are departed users.
For which project message about successful (or not) deletion will be printed.

1. Connect to the environment using data from workstation
2. Get all projects that the authenticated user has access to
3. Get all disabled users
4. For each project for each user get the list of subscriptions and delete them
5. For each deletion process message is displayed informing about the result
"""

from mstrio.api.projects import get_projects
from mstrio.connection import Connection, get_connection
from mstrio.distribution_services import SubscriptionManager
from mstrio.users_and_groups import list_users


def delete_subscriptions_of_departed_users(connection: "Connection") -> None:
    """Delete all subscription in all projects which owners are departed users.

    Args:
        Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
    """

    # get all projects that the authenticated user has access to
    response = get_projects(connection, whitelist=[('ERR014', 403)])
    projects_ = response.json() if response.ok else []
    # get all disabled users
    all_users = list_users(connection=connection)
    disabled_users = [u for u in all_users if not u.enabled]

    for project_ in projects_:
        project_id = project_['id']
        sub_manager = SubscriptionManager(connection=connection, project_id=project_id)
        for user_ in disabled_users:
            subs = sub_manager.list_subscriptions(owner={'id': user_.id})
            msg = f"subscriptions of user with ID: {user_.id}"
            msg += f" in project {project_['name']} with ID: {project_['id']}"
            # call of the function below returns True if all passed
            # subscriptions were deleted
            if sub_manager.delete(subscriptions=subs, force=True):
                print("All " + msg + " were deleted.", flush=True)
            else:
                print(
                    "Not all " + msg + " were deleted or there was no subscriptions.",
                    flush=True,
                )


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# execute deletion of addresses from departed users
delete_subscriptions_of_departed_users(conn)



from mstrio.api import projects as projects_api
from mstrio.connection import Connection


def _prepare_rest_output_for_project_languages(response: dict, path: str) -> dict:
    default_key = 'defaultLanguage' if path == 'data' else 'default'
    default_language_id = response.get(default_key)
    language = response.get('languages').get(default_language_id)
    response[default_key] = {default_language_id: language}
    return response


def get_project_languages(connection: Connection, id: str, path: str):
    """Get languages of the project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        path (string): Path to the languages.
            Available values:
                - `data`
                - `metadata`

    Returns:
        List of languages in form of a single dictionary.
    """
    return (
        projects_api.get_project_languages(connection=connection, id=id)
        .json()
        .get(path)
        .get('languages')
    )


def get_default_language(connection: Connection, id: str, path: str):
    """Get default language of the project.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project ID
        path (string): Path to the languages.
            Available values:
                - `data`
                - `metadata`
    """
    return _prepare_rest_output_for_project_languages(
        projects_api.get_project_languages(connection=connection, id=id)
        .json()
        .get(path),
        path,
    ).get('defaultLanguage' if path == 'data' else 'default')


def update_project_languages(connection: Connection, id: str, body: dict, path: str):
    return (
        projects_api.update_project_languages(connection=connection, id=id, body=body)
        .json()
        .get(path)
        .get('languages')
    )


def update_current_mode(connection: Connection, id: str, body: dict):
    return (
        projects_api.update_project_languages(connection=connection, id=id, body=body)
        .json()
        .get('data')
        .get('currentMode')
    )


def update_default_language(connection: Connection, id: str, body: dict):
    return _prepare_rest_output_for_project_languages(
        projects_api.update_project_languages(connection=connection, id=id, body=body)
        .json()
        .get('data'),
        path='data',
    ).get('defaultLanguage')


def get_project_internalization(connection: Connection, id: str):
    response = projects_api.get_project_languages(connection=connection, id=id).json()

    for path in ['data', 'metadata']:
        response[path] = _prepare_rest_output_for_project_languages(
            response=response[path], path=path
        )

    return response


def get_project_lock(connection: Connection, id: str):
    return {
        'lock_status': projects_api.get_project_lock(
            connection=connection, id=id
        ).json()
    }


def get_project_duplications(connection: Connection, limit: int | None) -> list:
    """Get project duplications."""
    limit_per_call = 1000
    offset = 0
    duplications = []

    while True:
        batch_limit = (
            min(limit_per_call, limit - len(duplications))
            if limit is not None
            else limit_per_call
        )
        new_duplications = (
            projects_api.get_project_duplications(
                connection=connection,
                limit=batch_limit,
                offset=offset,
            )
            .json()
            .get('duplications', [])
        )
        if len(new_duplications) == 0:
            break
        duplications.extend(new_duplications)
        offset += len(new_duplications)
        if limit is not None and len(duplications) >= limit:
            duplications = duplications[:limit]
            break
        if len(new_duplications) < batch_limit:
            break

    return duplications



import json
from collections.abc import Callable
from datetime import datetime
from enum import Enum
from functools import wraps

import pytz


class DatetimeFormats(Enum):
    FULLDATETIME = '%Y-%m-%dT%H:%M:%S.%f%z'
    DATE = '%Y-%m-%d'
    YMDHMSmS = '%Y-%m-%dT%H:%M:%S.%f%z'
    YMDHMS = '%Y-%m-%dT%H:%M:%S%z'
    YMD = '%Y-%m-%d'


def _without_loc(format_str):
    """Return format of date without localization."""
    if format_str.find('%z') != -1:
        return format_str[:-2]
    else:
        return format_str


def _adapt_date_to_format(date: str, format_str: str) -> tuple | str:
    """Adapt string with `date` to proper format provided in `format_str`.
    If time is incomplete, fill it with missing zeros to have format with
    milliseconds which then will be cut to match provided format. When
    localization is needed but it is not provided in `date` then `format_str`
    is changed not to require localization.

    Returns:
        Tuple `(date, format_str)` with new date and new format of date.
    """

    # save date to return it without changes in case of not supported date
    # format
    initial_date = date

    # save localization of date (if any) for later parsing
    if date[-1] == 'Z':
        date = date[:-1]
    plus_index = date.find('+')
    percent_z_index = format_str.find('%z')
    localization = (
        '' if plus_index == -1 or percent_z_index == -1 else date[plus_index:]
    )
    date = date[:plus_index] if plus_index != -1 else date

    # fill time with zeros to match the format with maximum number of details
    t_index = date.find('T')
    if t_index == -1:  # there is no time in the date
        date += 'T00:00:00.000'
        t_index = date.find('T')  # calculate again as this value was changed
    elif t_index + 3 == len(date):  # T00
        date += ':00:00.000'
    elif t_index + 6 == len(date):  # T00:00
        date += ':00.000'
    elif t_index + 9 == len(date):  # T00:00:00
        date += '.000'
    elif t_index + 13 < len(date):
        date = date[: t_index + 13]

    # cut `date` to match the provided format
    dot_index = date.find('.')

    if _without_loc(format_str) == _without_loc(DatetimeFormats.YMDHMSmS.value):
        # when `format_str` is equal to value of `DatetimeFormats.YMDHMSms`
        # then `date` is properly prepared and doesn't need any more changes
        pass
    elif _without_loc(format_str) == _without_loc(DatetimeFormats.YMDHMS.value):
        date = date[:dot_index]
    elif _without_loc(format_str) == _without_loc(DatetimeFormats.YMD.value):
        date = date[:t_index]
        localization = ''  # don't add localization if format is without time
    else:
        from mstrio.utils.helper import exception_handler

        msg = (
            f"For given format of date ({format_str}) adapting date to such format "
            f"is not provided."
        )
        exception_handler(msg, Warning)
        return (initial_date, format_str)

    # add localization prepared before if needed
    date += localization

    # cut localization from format string if it isn't present in date
    if percent_z_index != -1 and plus_index == -1:
        format_str = format_str[:-2]

    return (date, format_str)


def str_to_datetime(date: str, format_str: str) -> datetime | None:
    """Change date format to datetime, based on `format_str` provided.
    If `date` is already a datetime object, return it. Make the date aware."""
    if date is None or date == '':
        return date

    if not isinstance(date, datetime):
        date, format_str = _adapt_date_to_format(date, format_str)
        date = datetime.strptime(date, format_str)
    try:  # Localize to utc if not yet localized
        return pytz.utc.localize(date)
    except ValueError:  # Already localized
        return date


def _return_with_miliseconds(func):
    """If the date has milliseconds, return only 3 decimal places. Older servers
    need this to work. New can parse both 3 and 6 decimal places."""

    @wraps(func)
    def inner(*args, **kwargs):
        res = func(*args, **kwargs)
        if isinstance(res, str) and res.find('.') != -1:
            plus_pos = res.find('+')
            res = f"{res[:res.find('.') + 4]}{res[plus_pos:] if plus_pos != -1 else ''}"
        return res

    return inner


@_return_with_miliseconds
def datetime_to_str(date: datetime, format_str: str) -> str | None:
    """Get date string from datetime, based on `format_str` provided.
    If `date` is already a string, return it. Make the date aware."""
    if isinstance(date, str):
        return date
    try:
        # We need the date to be aware, or the string won't be accepted by API
        try:
            return pytz.utc.localize(date).strftime(format_str)
        except ValueError:  # Already localized
            return date.strftime(format_str)
    except (TypeError, AttributeError):
        return None


def _get_only_datetimeformat_map(string_to_date_map: dict) -> dict:
    """Return all entries that are of `DATETIMEFORMAT` enum type."""
    return {
        key: value
        for (key, value) in string_to_date_map.items()
        if isinstance(value, DatetimeFormats)
    }


def _solve_prefix_and_convert_date(
    func,
    name: str,
    date: str,
    string_to_date_map: dict,
    only_datetimefomat: bool = True,
):
    if only_datetimefomat:
        string_to_date_map = _get_only_datetimeformat_map(string_to_date_map)
    if f'_{name}' in string_to_date_map:
        date_format = (
            string_to_date_map[f'_{name}'].value
            if isinstance(string_to_date_map[f'_{name}'], DatetimeFormats)
            else string_to_date_map[f'_{name}']
        )
        return func(date, date_format)
    elif name in string_to_date_map:
        date_format = (
            string_to_date_map[name].value
            if isinstance(string_to_date_map[name], DatetimeFormats)
            else string_to_date_map[name]
        )
        return func(date, date_format)
    return date


def map_str_to_datetime(
    name: str, date: str, string_to_date_map: dict, only_datetimefomat: bool = True
) -> datetime:
    """Change date format to datetime, based on `string_to_date_map`
    conversion dict. All occurrences of `DATETIMEFORMAT` Enum in
    `string_to_date_map` are converted to corresponding string values.
    If name is not found in `string_to_date_map`, returns date without changes.
    """
    return _solve_prefix_and_convert_date(
        str_to_datetime, name, date, string_to_date_map, only_datetimefomat
    )


def map_datetime_to_str(
    name: str, date: datetime, string_to_date_map: dict, only_datetimefomat: bool = True
) -> str:
    """Change date format to string, based on `string_to_date_map`
    conversion dict. All occurrences of `DATETIMEFORMAT` Enum in
    `string_to_date_map` are converted to corresponding string values.
    If name is not found in `string_to_date_map`, returns date without changes.
    """
    return _solve_prefix_and_convert_date(
        datetime_to_str, name, date, string_to_date_map, only_datetimefomat
    )


def bulk_str_to_datetime(
    source: dict, string_to_date_map: dict, only_datetimefomat: bool = True
) -> dict:
    """Change all dates from `source` found in `string_to_date_map`
    to datetime format. If parameter is not found in `string_to_date_map`,
    it is returned without changes."""
    for key, val in source.items():
        source[key] = map_str_to_datetime(
            key, val, string_to_date_map, only_datetimefomat
        )
    return source


def bulk_datetime_to_str(
    source: dict, string_to_date_map: dict, only_datetimefomat: bool = True
) -> dict:
    """Change all dates from `source` found in `string_to_date_map`
    to string format. If parameter is not found in `string_to_date_map`,
    it is returned without changes."""
    for key, val in source.items():
        source[key] = map_datetime_to_str(
            key, val, string_to_date_map, only_datetimefomat
        )
    return source


def override_datetime_format(
    original_format: str, expected_format: str, fields: tuple, to_unpack=None
):
    """A decorator designed to override the datetime format
    of some dates in responses from REST server as they can be
    a bit crazy sometimes (e.g. two different formats for one object)

    Args:
        original_format: original format of a datetime
        expected_format: the format you want to convert to
        fields: fields of the object - e.g. dateModified, dateCreated
        to_unpack: when response returns a list of objects
            probably they need to be unpacked
    """

    def decorator_datetime(func: Callable):
        from mstrio.utils.helper import get_response_json

        @wraps(func)
        def wrapped(*args, **kwargs):
            response = func(*args, **kwargs)
            response_json = get_response_json(response)

            try:
                iterable = response_json[to_unpack] if to_unpack else [response_json]
            except KeyError:
                iterable = []

            for obj in iterable:
                for field in fields:
                    datetime_obj = str_to_datetime(obj[field], original_format)
                    obj[field] = datetime_to_str(datetime_obj, expected_format)

            response.encoding, response._content = 'utf-8', json.dumps(
                response_json
            ).encode('utf-8')
            return response

        return wrapped

    return decorator_datetime



import logging
from typing import TYPE_CHECKING, Any

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.server import Cluster
from mstrio.utils import helper
from mstrio.utils.version_helper import class_version_handler, meets_minimal_version

if TYPE_CHECKING:
    from mstrio.users_and_groups.user import User

logger = logging.getLogger(__name__)


@class_version_handler('11.2.0000')
class UserConnections:
    """Browse and manage active user connections on the environment. Use the
    `fetch()` method to fetch the latest active user connections. Filter the
    `user_connections` by using `filter_connections()` method.

    Attributes:
        connection: A Strategy One connection object
        user_connections: All active user connections on the environment
    """

    def __init__(self, connection: Connection):
        """Initialize the `UserConnections` object.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
        """
        self.connection = connection
        self.user_connections: list[dict[str, Any]] = []

    def fetch(self) -> None:
        """Populate the `UserConnections` object by retrieving all active user
        connections on the environment."""
        self.user_connections = self.list_connections()

    def filter_connections(self, **filters) -> list[dict[str, Any]] | None:
        """Filter the user connections stored in the `UserConnections` object
        by specifying the `filters` keyword arguments.

        Args:
            **filters: Available filter parameters: ['id', 'parent_id',
                'username', 'user_full_name', 'project_index', 'project_id',
                'project_name', 'open_jobs_count', 'project_type',
                'date_connection_created', 'duration', 'session_id', 'client',
                'config_level']
        """
        filtered_connections = None
        if not self.user_connections:
            helper.exception_handler(
                "Populate the `UserConnections` object with `UserConnections.fetch` "
                "first.",
                Warning,
            )
        else:
            filtered_connections = helper.filter_list_of_dicts(
                self.user_connections, **filters
            )
        return filtered_connections

    def list_connections(
        self,
        nodes: str | list[str] | None = None,
        limit: int | None = None,
        **filters,
    ) -> list[dict[str, Any]]:
        """Get all active user connections. Optionally filter the connections
        by specifying the `filters` keyword arguments.

        Args:
            nodes: Node (server) names on which users will be disconnected.
            limit: limit the number of elements returned. If `None`, all objects
                are returned.
            **filters: Available filter parameters: ['id', 'parent_id',
                'username', 'user_full_name', 'project_index', 'project_id',
                'project_name', 'open_jobs_count', 'project_type',
                'date_connection_created', 'duration', 'session_id', 'client',
                'config_level', 'application_type', 'last_action']
        """
        all_nodes = Cluster(self.connection).list_nodes(to_dictionary=True)
        all_connections = []
        if nodes is None:
            nodes = [node['name'] for node in all_nodes if node['status'] == 'running']
        else:
            nodes = nodes if isinstance(nodes, list) else [nodes]

        msg = 'Error fetching chunk of active user connections.'
        for node in nodes:
            all_connections.extend(
                helper.fetch_objects_async(
                    self.connection,
                    monitors.get_user_connections,
                    monitors.get_user_connections_async,
                    dict_unpack_value="userConnections",
                    limit=limit,
                    chunk_size=1000,
                    error_msg=msg,
                    node_name=node,
                    filters=filters,
                )
            )
        return all_connections

    def disconnect_users(
        self,
        connection_ids: str | list[str] | None = None,
        users: list["User"] | list[str] | None = None,
        nodes: str | list[str] | None = None,
        force: bool = False,
        **filters,
    ) -> list[dict] | None:
        """Disconnect user connections by passing in users (objects) or
        connection_ids. Optionally disconnect users by specifying the `filters`
        keyword arguments.

        Args:
            connection_ids: chosen ids that can be retrieved with
                `list_connections()`
            users: List of User objects or usernames
            nodes: Node (server) names on which users will be disconnected
            force: if True, no additional prompt will be shown before
                disconnecting users
            **filters: Available filter parameters: ['id', 'parent_id',
                'username', 'user_full_name', 'project_index', 'project_id',
                'project_name', 'open_jobs_count', 'project_type',
                'date_connection_created', 'duration', 'session_id', 'client',
                'config_level']
        Returns:
            - list of statuses of disconnecting chosen connections with theirs
              ids and messages from the I-Server:
                status code 200 is when all connections were disconnected
                status code 207 is when some connections were disconnected
                status code 403 without error code is when no connections were
                disconnected
            - in case of error of nothing to disconnect it returns None


        """
        # import here to avoid circular imports
        from mstrio.users_and_groups.user import User

        if (
            self.connection
            and not connection_ids
            and not users
            and not filters
            and not force
        ):
            msg = (
                "You need to pass connection_ids or users or specify filters. To "
                "disconnect all connections use `disconnect_all_users()` method."
            )
            helper.exception_handler(msg)

        if connection_ids:
            # disconnect specific user connections without fetching connections
            return self.__disconnect_by_connection_id(connection_ids)

        # get all user connections to filter locally
        all_connections = self.list_connections(nodes, **filters)
        if users:  # filter user connections by user objects
            users = users if isinstance(users, list) else [users]
            usernames = []
            for user in users:
                if isinstance(user, User):
                    usernames.append(user.username)
                elif isinstance(user, str):
                    usernames.append(user)
                else:
                    helper.exception_handler(
                        "'user' param must be a list of User objects or usernames.",
                        exception_type=TypeError,
                    )

            all_connections = list(
                filter(lambda conn: conn['username'] in usernames, all_connections)
            )

        if all_connections:
            # extract connection ids and disconnect
            connection_ids = [conn['id'] for conn in all_connections]
            return self.__disconnect_by_connection_id(connection_ids)
        elif config.verbose:
            logger.info('No active user connections.')

    def disconnect_all_users(self, force: bool = False) -> list[dict] | None:
        """Disconnect all user connections.

        Args:
            force: if True, no additional prompt will be shown before
                disconnecting all users

        Returns:
            - list of statuses of disconnecting all connections with their ids
               and messages from the I-Server
            - in case of error it returns None
        """
        if not force:
            user_input = input(
                "Are you sure you want to disconnect all users from the I-Server? "
                "[Y/N]: "
            )
            if user_input != "Y":
                return None
            else:
                force = True

        return self.disconnect_users(force=force)

    def __disconnect_by_connection_id(
        self, connection_ids: str | list[str]
    ) -> list[dict] | None:
        """It disconnects connections which ids are provided in
        'connection_ids'. It prints information about executed operations.
        Returns list of statuses of for the given connection ids with the
        messages from the I-Server or `None` in case of an error.
        """
        connection_ids = (
            connection_ids if isinstance(connection_ids, list) else [connection_ids]
        )
        server_version = (
            helper.version_cut(self.connection.iserver_version)
            if self.connection.iserver_version
            else None
        )

        # use monitors.delete_user_connections
        # or monitors.delete_user_connection depending on the server version
        if meets_minimal_version(server_version, '11.3.1'):
            res = monitors.delete_user_connections(
                connection=self.connection, ids=connection_ids
            )
            if res.status_code in [200, 207] or (
                res.status_code == 403 and not res.json().get('code')
            ):
                return self._prepare_disconnect_by_id_message(
                    statuses=res.json()['deleteUserConnectionsStatus']
                )
            else:
                err_msg = f'Error disconnecting user sessions: {connection_ids}.'
                return helper.response_handler(
                    response=res, msg=err_msg, throw_error=False
                )
        else:
            statuses: list[dict[str, str | int]] = []
            for connection_id in connection_ids:
                response = monitors.delete_user_connection(
                    self.connection, connection_id, bulk=True
                )
                statuses.append({'id': connection_id, 'status': response.status_code})
            return self._prepare_disconnect_by_id_message(statuses=statuses)

    @staticmethod
    def _prepare_disconnect_by_id_message(statuses: list[dict]) -> list[dict]:
        succeeded = []
        failed = []

        for s in statuses:
            status = s['status']
            id = s['id']
            if status >= 200 and status < 300:
                succeeded.append(id)
            elif status >= 400:
                # in most cases there will be status code 403
                failed.append(id)

        if config.verbose:
            if succeeded:
                logger.info(
                    'User connections with ids below were successfully disconnected:'
                    '\n\t' + ',\n\t'.join(succeeded)
                )
            if failed:
                logger.warning(
                    'User connections with ids below were not disconnected:\n\t'
                    + ',\n\t'.join(failed)
                )

        return statuses



"""
Export selected information about users and user groups to a CSV file.

1. Provide name of the project to connect to.
2. Connect to the environment using data from workstation.
3. Define filename where exported data will be saved.
4. Open the file and write the header row.
5. List all user groups and write their information to the file.
6. List all users and write their information to the file.
"""

from mstrio.connection import get_connection
from mstrio.users_and_groups import (
    list_user_groups,
    list_users,
)
import csv

# Define project to connect with
PROJECT_NAME = 'MicroStrategy Tutorial'  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Define filename where exported data will be saved. In Workstation you need to
# provide absolute path to the file remembering about double backslashes on WIN
# for example: 'C:\\Users\\admin\\Documents\\data\\users_and_user_groups.csv'
FILE_PATH = '<PATH_TO_FILE>'
with open(FILE_PATH, mode='w', newline='') as file:
    file_writer = csv.writer(
        file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL
    )
    file_writer.writerow(
        [
            'Full Name',
            'User ID',
            'Username (Login)',
            'Enabled',
            'Date Created',
            'Date Modified',
            'Description',
        ]
    )

    user_groups = list_user_groups(connection=conn)
    for user_group in user_groups:
        file_writer.writerow(
            [
                user_group.name,
                user_group.id,
                user_group.name,
                'N/A',
                user_group.date_created,
                user_group.date_modified,
                user_group.description,
            ]
        )

    users = list_users(connection=conn)
    for user in users:
        file_writer.writerow(
            [
                user.full_name,
                user.id,
                user.username,
                user.enabled,
                user.date_created,
                user.date_modified,
                user.description,
            ]
        )



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler

# These endpoints concern calendars, which are configuration-level objects.
# Project ID must be still provided. In this case an empty string is passed.


@unpack_information
@ErrorHandler(err_msg="Error getting calendars.")
def list_calendars(
    connection: Connection,
    subtype: str | None = None,
    offset: int | None = None,
    limit: int | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a list of calendars.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        subtype (str, optional): Subtype of the calendar to get. If skipped,
            all calendars are returned.
            Possible values: 'calendar_custom', 'calendar_system'
        offset (int, optional): Starting point within the collection of
            returned results. Used to control paging behavior.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Custom error message for error handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.get(
        endpoint='/api/model/calendars',
        headers={'X-MSTR-ProjectID': ''},
        params={
            'information.subType': subtype,
            'offset': offset,
            'limit': limit,
            'fields': fields,
        },
    )


@unpack_information
@ErrorHandler(err_msg="Error creating calendar.")
def create_calendar(
    connection: Connection,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Create a new calendar.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new calendar
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    with changeset_manager(
        connection, project_id='', schema_edit=False
    ) as changeset_id:
        return connection.post(
            endpoint='/api/model/calendars',
            json=body,
            headers={'X-MSTR-MS-Changeset': changeset_id},
        )


@unpack_information
@ErrorHandler(err_msg="Error getting calendar with ID: {id}")
def get_calendar(
    connection: Connection,
    id: str,
    changeset_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a calendar by its identifier.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): Identifier of the calendar to get
        changeset_id (str, optional): Identifier of the changeset to get the
            calendar from
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint=f'/api/model/calendars/{id}',
        headers={'X-MSTR-ProjectID': '', 'X-MSTR-MS-Changeset': changeset_id},
    )


@unpack_information
@ErrorHandler(err_msg="Error updating calendar with ID: {id}")
def update_calendar(
    connection: Connection,
    id: str,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Update an existing calendar.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): Identifier of the calendar to update
        body (dict): JSON-formatted body of the updated calendar
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    with changeset_manager(
        connection, project_id='', schema_edit=False
    ) as changeset_id:
        return connection.put(
            endpoint=f'/api/model/calendars/{id}',
            json=body,
            headers={'X-MSTR-MS-Changeset': changeset_id},
        )



from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from mstrio.utils.helper import Dictable
from mstrio.utils.time_helper import DatetimeFormats


class SubscriptionState(Enum):
    """State of the subscription"""

    INVALID = 0
    SUCCESS = 1
    FAIL = 2
    CANCELED = 3
    TIMEOUT = 4
    SKIP = 5


class SubscriptionStage(Enum):
    """Stage of the subscription"""

    INVALID = 0
    EXECUTING = 1
    FINISHED = 2


@dataclass
class StatusError(Dictable):
    """Subscription status error

    Attributes:
        code (int): Error code.
        message (str): Error message.
    """

    code: int
    message: str


@dataclass
class StatusContent(Dictable):
    """Subscription status content

    Attributes:
        name (str): Content name
        node_name (str): Node name
        error (StatusError): Error
    """

    _FROM_DICT_MAP = {
        'error': lambda source, connection: (StatusError.from_dict(source, connection)),
    }

    name: str
    node_name: str | None = None
    error: StatusError | None = None


@dataclass
class StatusDestination(Dictable):
    """Subscription status destination

    Attributes:
        address (str): Address
        destination (str): Destination
        start (datetime): Start time
        end (datetime): End time
        node_name (str): Node name
        contents (list[StatusContent]): Subscription contents
        error (StatusError): Error
    """

    _FROM_DICT_MAP = {
        'start': DatetimeFormats.FULLDATETIME,
        'end': DatetimeFormats.FULLDATETIME,
        'contents': lambda source, connection: [
            StatusContent.from_dict(content, connection) for content in source
        ],
        'error': lambda source, connection: (StatusError.from_dict(source, connection)),
    }

    address: str
    destination: str
    start: datetime
    node_name: str
    end: datetime | None = None
    contents: list[StatusContent] | None = None
    error: StatusError | None = None


@dataclass
class StatusDetail(Dictable):
    """Subscription status detail

    Attributes:
        recipient (str): Recipient
        contents (list[StatusContent]): Subscription contents
        destinations (list[StatusDestination]): Destinations
        error (StatusError): Error
    """

    _FROM_DICT_MAP = {
        'contents': lambda source, connection: [
            StatusContent.from_dict(content, connection) for content in source
        ],
        'destinations': lambda source, connection: [
            StatusDestination.from_dict(content, connection) for content in source
        ],
        'error': lambda source, connection: (StatusError.from_dict(source, connection)),
    }

    recipient: str
    contents: list[StatusContent]
    destinations: list[StatusDestination]
    error: StatusError | None = None


@dataclass
class SubscriptionStatus(Dictable):
    """Subscription status

    Attributes:
        id (str): Subscription ID
        stage (SubscriptionStage): Stage of the subscription
        state (SubscriptionState): State of the subscription
        total (int): Total number of statuses
        estimate (int): Estimated time remaining in seconds
        start (datetime): Subscription start time
        end (datetime): Subscription end time
        contents (list[StatusContent]): Subscription contents
        statuses (list[StatusDetail]): Subscription detailed statuses
        expiration (str): Subscription expiration time
    """

    _FROM_DICT_MAP = {
        'contents': lambda source, connection: [
            StatusContent.from_dict(content, connection) for content in source
        ],
        'start': DatetimeFormats.FULLDATETIME,
        'end': DatetimeFormats.FULLDATETIME,
        'stage': SubscriptionStage,
        'state': SubscriptionState,
        'statuses': lambda source, connection: [
            StatusDetail.from_dict(content, connection) for content in source
        ],
    }

    id: str
    stage: SubscriptionStage
    state: SubscriptionState
    total: int
    estimate: int
    start: datetime
    contents: list[StatusContent]
    statuses: list[StatusDetail]
    end: datetime | None = None
    expiration: str | None = None



from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting Gateways.")
def get_gateways(
    connection: Connection,
    error_msg: str | None = None,
):
    """Get information for all gateways.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """

    return connection.get(
        endpoint='/api/gateways',
    )


@ErrorHandler(err_msg='Error getting Gateway with ID {id}')
def get_gateway(connection: Connection, id: str, error_msg: str | None = None):
    """Get gateway by a specific ID.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the gateway
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    return connection.get(
        endpoint=f'/api/gateways/{id}',
    )



"""This is the demo script to show how to manage agents. Its basic goal
is to present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.agents import Agent, list_agents

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

# Get connection to the environment
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# List all agents
agents = list_agents(connection=conn)
# List all agents as dictionaries
agents_as_dicts = list_agents(connection=conn, to_dictionary=True)

# Define a variable which can be later used in a script
AGENT_ID = $agent_id

# Initialize an agent
test_agent = Agent(connection=conn, id=AGENT_ID)

# Define variables which can be later used in a script
NEW_NAME = $new_name
NEW_DESCRIPTION = $new_description
NEW_ABBREVIATION = $new_abbreviation
NEW_HIDDEN = $new_hidden == 'True'
NEW_STATUS = $new_status
NEW_COMMENTS = $new_comments

# Alter the agent's properties
test_agent.alter(
    name=NEW_NAME,
    description=NEW_DESCRIPTION,
    abbreviation=NEW_ABBREVIATION,
    hidden=NEW_HIDDEN,
    status=NEW_STATUS,
    comments=NEW_COMMENTS,
)

# Disable and enable an agent
test_agent.disable()
test_agent.enable()

# Delete an agent
test_agent.delete(force=True)



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler

# These endpoints concern timezones, which are configuration-level objects.
# Project ID must be still provided. In this case an empty string is passed.


@unpack_information
@ErrorHandler(err_msg="Error getting timezones.")
def list_all_tzs(
    connection: Connection,
    subtype: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a list of timezones.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        subtype (str, optional): Subtype of the timezone to get. If skipped,
            all timezones are returned.
            Possible values: 'timezone_custom', 'timezone_system'
        error_msg (str, optional): Custom error message for error handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.get(
        endpoint='/api/model/timezones',
        headers={'X-MSTR-ProjectID': ''},
        params={'subType': subtype},
    )


@unpack_information
@ErrorHandler(err_msg="Error creating timezone.")
def create_tz(
    connection: Connection,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Create a new timezone.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new timezone
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    with changeset_manager(
        connection, project_id='', schema_edit=False
    ) as changeset_id:
        return connection.post(
            endpoint='/api/model/timezones',
            json=body,
            headers={'X-MSTR-MS-Changeset': changeset_id},
        )


@unpack_information
@ErrorHandler(err_msg="Error getting timezone with ID: {id}")
def get_tz(
    connection: Connection,
    id: str,
    changeset_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a timezone by its identifier.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): Identifier of the timezone to get
        changeset_id (str, optional): Identifier of the changeset to get the
            timezone from
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint=f'/api/model/timezones/{id}',
        headers={'X-MSTR-ProjectID': '', 'X-MSTR-MS-Changeset': changeset_id},
    )


@unpack_information
@ErrorHandler(err_msg="Error updating timezone with ID: {id}")
def update_tz(
    connection: Connection,
    id: str,
    body: dict,
    error_msg: str | None = None,
) -> Response:
    """Update an existing timezone.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): Identifier of the timezone to update
        body (dict): JSON-formatted body of the updated timezone
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    with changeset_manager(
        connection, project_id='', schema_edit=False
    ) as changeset_id:
        # Note: This endpoint uses the PATCH method, but the request body has
        # the simple delta format, not the patch operation list format
        return connection.patch(
            endpoint=f'/api/model/timezones/{id}',
            json=body,
            headers={'X-MSTR-MS-Changeset': changeset_id},
        )



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.helper import Dictable

from .expression import Expression

if TYPE_CHECKING:
    from mstrio.connection import Connection


@dataclass
class FactExpression(Dictable):
    """Object that gives information about an expression that is in use by
    a fact or an attribute.
    Naturally the most important part of a fact expression is the expression.
    This is often just a single column reference, but more complex expressions
    are supported. The object also lists those tables where the expression is
    currently used or where it could be used.

    We say that a table supports a fact if it is possible to compute a value
    for the fact by means of a SELECT expression for each row of the table. The
    expression represents a SELECT expression.

    Args:

        id (str): FactExpression's id, a globally unique identifier used to
        distinguish between metadata objects within the same project.

        expression (Expression): A generic specification for a calculation
        stored within a metadata object. The expression is represented as a
        tree over nodes. Most internal nodes (called operator nodes) are
        defined by applying a function to the operator's child nodes.

        Usually an expression must be non-empty. But in a few cases, most
        notably a filter expression it is valid for an expression to contain no
        nodes at all. An expression is empty if and only if it does not have a
        tree property.

        Alternatively the client may prefer to handle an expression as a list
        of tokens. Each token represents part of the raw text of the
        expression, in some cases annotated with additional information.

        tables (List[SchemaObjectReference]): Array with a member object for
        each logical table that computes a value for this fact by evaluating
        the expression.

        This array should be non-empty for a fact or an attribute that is saved
        into metadata, but may be empty during an edit changeset. The order of
        the tables within the array has no significance.
    """

    _FROM_DICT_MAP = {
        'tables': [SchemaObjectReference],
        'expression': Expression,
    }

    expression: Expression
    tables: list[SchemaObjectReference]
    id: str | None = None

    def to_dict(self, camel_case: bool = True) -> dict:
        result = super().to_dict(camel_case)
        result['expressionId'] = result.pop('id', None)

        return result

    @classmethod
    def from_dict(
        cls,
        source: dict,
        connection: Optional['Connection'] = None,
        to_snake_case: bool = True,
    ) -> 'FactExpression':
        data = source.copy()
        data['id'] = data.get('expressionId')

        return super().from_dict(data, connection, to_snake_case)

    def local_alter(
        self, expression: Expression = None, tables: list[SchemaObjectReference] = None
    ):
        """Locally alters the Fact Expression

        Args:
            expression: Expression object that will replace the current
                Expression representing the calculations required for this fact
            tables: array of schema object references that will replace the
                current array with logical tables needed to represent
                this fact"""
        if expression:
            self.expression = expression

        if tables:
            self.tables = tables



"""
Manage access control lists (ACLs).
Assign ACLs to user groups and its children for folders.

1. Connect to the environment using data from workstation
2. Assign ACL with View aggregated rights to user group with name 'Mobile' for
   user 'Administrator'
3. Assign ACL with View aggregated rights to user group with name 'Mobile' for
   user 'Administrator' with propagation to children of this user group
4. Assign ACL with every right except Execution to folder with given id for user
   'Administrator'
5. Assign ACL with every right except Execution to folder with given id for user
   'Administrator' with propagation to children of this folder
"""

from mstrio.connection import get_connection
from mstrio.object_management.folder import Folder
from mstrio.helpers import AggregatedRights
from mstrio.users_and_groups.user import User
from mstrio.users_and_groups.user_group import list_user_groups

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# assign ACL AggregatedRights.VIEW to UserGroup
user_group = list_user_groups(conn, name='Mobile')[0]
user_group.acl_add(
    rights=AggregatedRights.VIEW,
    trustees=[User(conn, name='Administrator')],
    denied=False,
)

# assign ACL AggregatedRights.VIEW to UserGroup and propagate it to children
user_group = list_user_groups(conn, name='Mobile')[0]
user_group.acl_add(
    rights=AggregatedRights.VIEW,
    trustees=[User(conn, name='Administrator')],
    denied=False,
    propagate_to_children=True,
)

# assign ACLs for folder
folder = Folder(conn, id='945A968E40E4EA2C9114BBBE885DBC24')
folder.acl_add(rights=127, trustees=[User(conn, name='Administrator')], denied=False)

# assign ACLs for folder and propagate to children
folder = Folder(conn, id='945A968E40E4EA2C9114BBBE885DBC24')
folder.acl_add(
    rights=127,
    trustees=[User(conn, name='Administrator')],
    denied=False,
    propagate_to_children=True,
)



import inspect
import logging
from functools import wraps
from inspect import getattr_static, getmembers

from packaging.version import Version as PackageVersion
from packaging.version import parse as ver_parser

from mstrio.connection import Connection
from mstrio.helpers import VersionException

logger = logging.getLogger(__name__)


def meets_minimal_version(
    tested_version: str | PackageVersion | None,
    minimal_version: str | PackageVersion,
    throw_error: bool = False,
) -> bool:
    """
    Checks if tested version is greater or equal to minimal version.

    Args:
        tested_version (str | PackageVersion | None): Tested version.
        minimal_version (str | PackageVersion): Minimal version.
        throw_error (bool, optional): If True, will raise VersionException
            if tested version is None. Defaults to False.

    Returns:
        bool: True if tested version is greater or equal to minimal version.
              If tested version is `None`, True is returned as we assume
              we were not able to gather value to test.
    """
    if tested_version is None:
        if throw_error:
            raise VersionException("Version to compare not provided.")

        logger.warning(
            "Version to compare not provided. "
            f"Assuming minimal version '{minimal_version}' is met."
        )
        return True

    to_test = (
        tested_version
        if isinstance(tested_version, PackageVersion)
        else ver_parser(tested_version)
    )
    minimal = (
        minimal_version
        if isinstance(minimal_version, PackageVersion)
        else ver_parser(minimal_version)
    )

    return to_test >= minimal


def method_version_handler(version):
    """
    Decorator which can be applied to a function.
    When applied, validates the iServer version of the existing connection
    (retrieved from kwargs, args or `connection` property from self)
    and compares it with the `version` provided in the arguments -
    version at which the functionality becomes applicable.
    """

    def wrapper(function, cls=None):
        @wraps(function)
        def inner(*args, **kwargs):
            if conn := kwargs.get('connection'):
                connection_obj = conn
            elif conn := [arg for arg in args if isinstance(arg, Connection)]:
                connection_obj = conn[0]
            elif cls:
                raise TypeError(
                    f"{function.__name__}() " "missing required argument: 'connection'"
                )
            else:
                connection_obj = getattr(args[0], 'connection', None) or getattr(
                    args[0], '_connection', None
                )

            if not meets_minimal_version(connection_obj.iserver_version, version):
                raise VersionException(
                    f"Environments must run IServer version {version} or newer. "
                    "Please update your environments to use this feature."
                )
            return function(*args, **kwargs)

        return inner

    return wrapper


def class_version_handler(version):
    """
    Decorator which can be used on a class.
    It applies the `method_version_handler decorator` to all the methods
    in the decorated class that are not inherited.
    """
    decorator = method_version_handler(version)

    def wrapper(cls):
        for name, obj in getmembers(cls):
            if name in vars(cls):
                # get attribute without triggering descriptor protocol
                attr = getattr_static(cls, name)

                if inspect.isfunction(attr) and name == '__init__':
                    # if __init__ method
                    setattr(cls, name, decorator(obj, cls))
                elif inspect.isfunction(attr):
                    # if other instance method
                    setattr(cls, name, decorator(obj))
                elif (
                    isinstance(attr, staticmethod)
                    and Connection in obj.__annotations__.values()
                ):
                    # if static method with Connection in params
                    setattr(
                        cls,
                        name,
                        staticmethod(decorator(obj, cls=cls)),
                    )
                elif (
                    isinstance(attr, classmethod)
                    and Connection in obj.__annotations__.values()
                ):
                    # if class method with Connection in params
                    setattr(
                        cls,
                        name,
                        classmethod(decorator(obj.__func__, cls=cls)),
                    )
                # if static or class method without params of type Connection
                # -> skip it

        return cls

    return wrapper


def is_server_min_version(connection: 'Connection', version_str: str) -> bool:
    """Check if iServer version is greater or equal than given version.

    Args:
        connection (Connection): Strategy One REST API connection object
        version_str (str): String containing iServer version number

    Returns:
        True if iServer version is greater or equal to given version
            or if it cannot be checked.
        False if iServer version is lower than given version.
    """
    return meets_minimal_version(connection.iserver_version, version_str)



import logging
from enum import Enum, auto
from typing import TYPE_CHECKING, Union

from packaging import version

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.helpers import (
    IServerError,
    MstrException,
    PartialSuccess,
    Success,
    VersionException,
)
from mstrio.server import Node, Project
from mstrio.utils.entity import Entity, EntityBase
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import exception_handler, validate_param_value
from mstrio.utils.monitors import all_nodes_async
from mstrio.utils.time_helper import DatetimeFormats, map_str_to_datetime
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.users_and_groups import User

logger = logging.getLogger(__name__)

ISERVER_VERSION_11_3_2 = '11.3.0200'


class DeliveryType(AutoName):
    RESERVED = auto()
    EMAIL = auto()
    FILE = auto()
    PRINTER = auto()
    CUSTOM = auto()
    INBOX = auto()
    CLIENT = auto()
    CACHE = auto()
    MOBILE = auto()
    BLACKBERRY = auto()
    IPHONE = auto()
    IPAD = auto()
    CAMPAIGN = auto()
    SNAPSHOT = auto()
    FTP = auto()
    ANDROID = auto()
    LAST_ONE = auto()
    COUNT = auto()
    ALL_INCLUDING_SNAPSHOT = auto()


class SortByV1(Enum):
    JOB_TYPE = 'jobType'
    USER_FULL_NAME = 'userFullName'
    OBJECT_ID = 'objectId'
    STATUS = 'status'
    PROJECT_ID = 'projectId'
    # check if valid as there are multiple
    # inconsistent information about sorting
    PROJECT_NAME = 'projectName'
    CREATION_TIME = 'creationTime'
    DESCRIPTION = 'description'
    JOB_ID = 'jobId'


class JobStatus(AutoName):
    READY = auto()
    EXECUTING = auto()
    WAITING = auto()
    COMPLETED = auto()
    ERROR = auto()
    CANCELING = auto()
    STOPPED = auto()
    WAITING_ON_GOVERNOR = auto()
    WAITING_FOR_AUTOPROMPT = auto()
    WAITING_FOR_PROJECT = auto()
    WAITING_FOR_CACHE = auto()
    WAITING_FOR_CHILDREN = auto()
    WAITING_FOR_RESULTS = auto()
    LOADING_PROMPT = auto()
    RESOLVING_DESTINATION = auto()
    DELIVERING = auto()
    EXPORTING = auto()
    CACHE_READY = auto()
    WAITING_FOR_DI_FILE = auto()
    WAITING_FOR_CONFLICT_RESOLVE = auto()
    STEP_PAUSING = auto()


class JobType(AutoName):
    INTERACTIVE = auto()
    SUBSCRIPTION = auto()
    PREDICTIVE_CACHE = auto()


# NOTE: new endpoint Enums below:
class ObjectType(AutoName):
    OTHERS = auto()
    REPORT = auto()
    CUBE = auto()
    DOCUMENT = auto()
    DASHBOARD = auto()

    @classmethod
    def _missing_(cls, value):
        """Handle missing values in ObjectType enum."""
        if value == 'dossier':
            return cls.DASHBOARD
        else:
            return super()._missing_(value)


class PUName(AutoName):
    BROWSING = auto()
    RESOLUTION = auto()
    QUERY_EXECUTIONS = 'query_execution'
    ANALYTICAL = auto()
    SQL_ENGINE = auto()
    DATA_FORMATTING = auto()
    NCS = auto()
    REST_ASYNCHRONOUS = auto()


class SubscriptionType(AutoName):
    RESERVED = auto()
    EMAIL = auto()
    PRINTER = auto()
    CUSTOM = auto()
    INBOX = auto()
    CLIENT = auto()
    CACHE = auto()
    MOBILE = auto()
    MOBILE_BLACKBERRY = auto()
    MOBILE_IPHONE = auto()
    MOBILE_IPAD = auto()
    CAMPAIGN = auto()
    SNAPSHOT = auto()
    FTP = auto()
    MOBILE_ANDROID = auto()
    LAST_ONE = auto()
    COUNT = auto()
    ALL = auto()
    ALL_INCLUDING_SNAPSHOT = auto()


class SortBy(Enum):
    ID_ASC = '+id'
    ID_DESC = '-id'
    TYPE_ASC = '+type'
    TYPE_DESC = '-type'
    STATUS_ASC = '+status'
    STATUS_DESC = '-status'
    USER_ASC = '+user'
    USER_DESC = '-user'
    DESCRIPTION_ASC = '+description'
    DESCRIPTION_DESC = '-description'
    OBJECT_TYPE_ASC = '+objectType'
    OBJECT_TYPE_DESC = '-objectType'
    OBJECT_ID_ASC = '+objectId'
    OBJECT_ID_DESC = '-objectId'
    SUBSCRIPTION_TYPE_ASC = '+subscriptionType'
    SUBSCRIPTION_TYPE_DESC = '-subscriptionType'
    PROCESSING_UNIT_PRIORITY_ASC = '+processingUnitPriority'
    PROCESSING_UNIT_PRIORITY_DESC = '-processingUnitPriority'
    CREATION_TIME_ASC = '+creationTime'
    CREATION_TIME_DESC = '-creationTime'
    COMPLETED_TASKS_ASC = '+completedTasks'
    COMPLETED_TASKS_DESC = '-completedTasks'
    PROJECT_ID_ASC = '+projectId'
    PROJECT_ID_DESC = '-projectId'
    PROJECT_NAME_ASC = '+projectName'
    PROJECT_NAME_DESC = '-projectName'
    SUBSCRIPTION_RECIPIENT_ASC = '+subscriptionRecipient'
    SUBSCRIPTION_RECIPIENT_DESC = '-subscriptionRecipient'
    MEMORY_USAGE_ASC = '+memoryUsage'
    MEMORY_USAGE_DESC = '-memoryUsage'
    ELAPSED_TIME_ASC = '+elapsedTime'
    ELAPSED_TIME_DESC = '-elapsedTime'


@method_version_handler('11.3.0200')
def _set_api_wrappers(connection) -> dict:
    if connection.iserver_version and version.parse(
        connection.iserver_version
    ) == version.parse(ISERVER_VERSION_11_3_2):
        return {
            (
                'id',
                'description',
                'status',
                'type',
                'priority',
                'processing_unit_priority',
                'warehouse_priority',
                'object_id',
                'user',
                'project_name',
                'creation_time',
                'total_tasks',
                'completed_tasks',
                'filter_name',
                'template_name',
                'sql',
                'subscription_owner',
                'subscription_recipient',
                'destination',
            ): monitors.get_job
        }
    else:
        return {
            (
                'id',
                'type',
                'status',
                'user',
                'description',
                'object_type',
                'object_id',
                'parent_id',
                'child_ids',
                'subscription_type',
                'processing_unit_priority',
                'step_id',
                'pu_name',
                'creation_time',
                'completed_tasks',
                'total_tasks',
                'project_name',
                'subscription_recipient',
                'memory_usage',
                'elapsed_time',
                'step_elapsed_time',
                'filter_name',
                'template_name',
                'sql',
                'subscription_owner',
                'error_time',
                'error_message',
                'step_statistics',
            ): monitors.get_job_v2
        }


@method_version_handler('11.3.0200')
def list_jobs(
    connection: "Connection",
    node: Node | str | None = None,
    user: Union["User", str] | None = None,
    description: str | None = None,
    type: JobType | str | None = None,
    status: JobStatus | str | None = None,
    object_id: str | None = None,
    object_type: ObjectType | None = None,
    project: Project | str | None = None,
    pu_name: PUName | str | None = None,
    subscription_type: SubscriptionType | str | None = None,
    subscription_recipient: Union["User", str] | None = None,
    memory_usage: str | None = None,
    elapsed_time: str | None = None,
    sort_by: SortBy | str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Job"] | list[dict]:
    """List jobs objects or job dictionaires.
    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`
        node(Node, str, optional): Node object or name, if not passed list jobs
            on all nodes
        user(User, str, optional): Field to filter on job owner's full name
        (exact match)
        description(str, optional): Field to filter on job description (partial
            match)
        type(JobType, optional): Field to filter on job type (exact match),
        status(JobStatus, optional): Field to filter on job status (exact match)
        object_id(str, optional): Field to filter on object type (exact
            match)
        object_type(ObjectType, optional): Field to filter on object type (exact
            match)
        project(Project, str, optional): Field to filter on project name (exact
            match)
        pu_name(PUName, optional): Field to filter on processing unit name
            (exact match)
        subscription_type(SubscriptionType, optional): Field to filter on
        subscription type (exact match)
        subscription_recipient(User, str, optional): Field to filter on
            subscription recipient's full name (exact match)
        memory_usage(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with memory usage greater
            than to 100 MB. Valid operators are:
            gt - greater than
            lt - less than
        elapsed_time(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with elapsed time greater
            than 100 milliseconds(11.3.3+) or seconds(11.3.2). Valid operators:
            gt - greater than
            lt - less than
        sort_by(SortBy, optional): Specify sorting criteria, for example
            SortBy.STATUS_ASC means sorting status is ascending order
            or SortBy.USER_DESC means sorting user in descending order.
            Currently, the server supports sorting only by single field.
        to_dictionary(bool, optional): if True, return Schedules as
            list of dicts
        limit(int, optional): maximum number of schedules returned
        **filters(optional): Local filter parameters

    Examples:
        >>> list_jobs_v1(connection, duration='gt:100')

    Returns:
        List[Job] | List[dict]: list of Job objects or dictionaries.
    """
    user = user.full_name if isinstance(user, Entity) else user
    subscription_recipient = (
        subscription_recipient.full_name
        if isinstance(user, Entity)
        else subscription_recipient
    )
    if project:
        project = (
            project
            if isinstance(project, Project)
            else Project(connection, name=project)
        )

    if isinstance(object_type, str) and object_type.lower() == 'dashboard':
        object_type = ObjectType.DASHBOARD

    # depending on version call either one or the other
    if connection.iserver_version and version.parse(
        connection.iserver_version
    ) == version.parse(ISERVER_VERSION_11_3_2):
        filters = __prepare_v1_request(
            description,
            object_type,
            pu_name,
            subscription_type,
            subscription_recipient,
            memory_usage,
            elapsed_time,
            filters,
        )
        return list_jobs_v1(
            connection=connection,
            node=node,
            project=project,
            status=status,
            job_type=type,
            user=user,
            limit=limit,
            to_dictionary=to_dictionary,
            **filters,
        )
    else:
        project_name = project.name if project else None
        node_name = node.name if isinstance(node, Node) else node
        msg = "Error fetching chunk of jobs."
        objects = all_nodes_async(
            connection,
            async_api=monitors.get_jobs_v2_async,
            filters=filters,
            error_msg=msg,
            unpack_value='jobs',
            limit=limit,
            node_name=node_name,
            user=user,
            description=description,
            type=type,
            object_id=object_id,
            object_type=object_type,
            project_name=project_name,
            status=status,
            pu_name=pu_name,
            subscription_type=subscription_type,
            subscription_recipient=subscription_recipient,
            memory_usage=memory_usage,
            elapsed_time=elapsed_time,
            sort_by=sort_by,
            fields=['jobs'],
        )

    if to_dictionary:
        return objects
    else:
        return [Job.from_dict(source=obj, connection=connection) for obj in objects]


@method_version_handler('11.3.0200')
def list_jobs_v1(
    connection: Connection,
    node: Node | str | None = None,
    project: Project | str | None = None,
    status: JobStatus | str | None = None,
    job_type: JobType | str | None = None,
    user: Union["User", str] | None = None,
    object_id: str | None = None,
    sort_by: SortBy | str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["Job"] | list[dict]:
    """List job objects or job dictionaries. Optionally filter list.
    NOTE: list_jobs can return up to 1024 jobs per request.

    Args:
        connection(object): Strategy One connection object returned by
            'connection.Connection()'
        node(Node, str, optional): Node object or name, if not passed list jobs
            on all nodes
        project(str, Project, optional): Project id or object to filter by
        status(JobStatus, optional): Job status to filter by
        job_type(JobType, optional): Job type to filter by
        user(User, str, optional): User object or full name to filter by
        object_id(str, optional): Object id to filter by
        sort_by(SortBy, optional): Specifies sorting criteria to sort by
        to_dictionary(bool, optional): if True, return Schedules as
            list of dicts
        limit(int, optional): maximum number of schedules returned.
        **filters: Available filter parameters:[id, description, status,
            jobType, duration, jobId, objectd]
            - duration(str, optional): Field to filter on the job elapsed time,
                for example 'gt:100' means filtering jobs with elapsed time
                greater than to 100 seconds. Valid operators:
                gt - greater than
                lt - less than

    Examples:
        >>> list_jobs_v1(connection, duration='gt:100')

    Returns:
        List[Job] | List[dict]: list of Job objects or dictionaries.
    """
    project_id = project.id if isinstance(project, Project) else project
    user_full_name = user.full_name if isinstance(user, Entity) else user

    if 'duration' in filters:
        filters['duration'] = __elapsed_filtering(filters['duration'])

    node_name = node.name if isinstance(node, Node) else node
    msg = "Error fetching chunk of jobs."
    objects = all_nodes_async(
        connection,
        async_api=monitors.get_jobs_async,
        filters=filters,
        error_msg=msg,
        unpack_value='jobs',
        limit=limit,
        node_name=node_name,
        project_id=project_id,
        status=status,
        job_type=job_type,
        user_full_name=user_full_name,
        object_id=object_id,
        sort_by=sort_by,
    )

    if to_dictionary:
        return objects
    else:
        return [Job.from_dict(source=obj, connection=connection) for obj in objects]


@method_version_handler('11.3.0200')
def kill_jobs(
    connection: Connection, jobs: list[Union["Job", str]]
) -> Success | PartialSuccess | MstrException:
    """Kill existing jobs by Job objects or job ids.

    Args:
        connection(object): Strategy One connection object returned by
            'connection.Connection()'
        jobs: List of Job objects or job ids to kill

    Returns:
        Success: object if all jobs were killed
        PartialSuccess: if not all jobs were killed
        MstrException: otherwise
    """
    validate_param_value("jobs", jobs, list)
    jobs = [job.id if isinstance(job, Job) else job for job in jobs]
    return monitors.cancel_jobs(connection, jobs)


@method_version_handler('11.3.0200')
def kill_all_jobs(
    connection: Connection,
    user: Union["User", str] | None = None,
    description: str | None = None,
    type: JobType | str | None = None,
    status: JobStatus | str | None = None,
    object_id: str | None = None,
    object_type: ObjectType | str | None = None,
    project: Project | str | None = None,
    pu_name: PUName | str | None = None,
    subscription_type: SubscriptionType | str | None = None,
    subscription_recipient: Union["User", str] | None = None,
    memory_usage: str | None = None,
    elapsed_time: str | None = None,
    force: bool = False,
    **filters,
) -> Success | PartialSuccess | MstrException:
    """Kill jobs filtered by passed fields

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`
         node(Node, str, optional): Node object or name, if not passed kill jobs
            on all nodes
        user(User, str, optional): Field to filter on job owner's full name
        (exact match)
        description(str, optional): Field to filter on job description (partial
            match)
        type(JobType, optional): Field to filter on job type (exact match),
        status(JobStatus, optional): Field to filter on job status (exact match)
        object_id(str, optional): Field to filter on object id (exact
            match)
        object_type(ObjectType, optional): Field to filter on object type (exact
            match)
        project(Project, str, optional): Field to filter on project name (exact
            match)
        pu_name(PUName, optional): Field to filter on processing unit name
            (exact match)
        subscription_type(SubscriptionType, optional): Field to filter on
        subscription type (exact match)
        subscription_recipient(User, str, optional): Field to filter on
            subscription recipient's full name (exact match)
        memory_usage(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with memory usage greater
            than to 100 MB. Valid operators are:
            gt - greater than
            lt - less than
        elapsed_time(str, optional): Field to filter on the job elapsed time,
            for example 'gt:100' means filtering jobs with elapsed time greater
            than 100 milliseconds(11.3.3+) or seconds(11.3.2). Valid operators:
            gt - greater than
            lt - less than
        force(bool, optional): Force flag

    Returns:
        Success: object if all jobs were killed
        PartialSuccess: if not all jobs were killed
        MstrException: otherwise
    """

    jobs = list_jobs(
        connection,
        user,
        description,
        type,
        status,
        object_id,
        object_type,
        project,
        pu_name,
        subscription_type,
        subscription_recipient,
        memory_usage,
        elapsed_time,
        filters,
    )
    assert jobs, "No jobs to kill"

    jobs_ids = [job.id for job in jobs]
    if not force:
        print(f"Found jobs: {jobs}")
        user_input = input("Are you sure you want to kill those jobs?[Y/N]: ") or 'N'
    if force or user_input == 'Y':
        return kill_jobs(connection, jobs_ids)


def __prepare_v1_request(
    description: str,
    object_type: str,
    pu_name: str,
    subscription_type: str,
    subscription_recipient: str,
    memory_usage: str,
    elapsed_time: str,
    filters: dict,
) -> dict:
    # raise VersionException if parameter not supported in v1
    unsupported_parameters = [
        description,
        object_type,
        pu_name,
        subscription_type,
        subscription_recipient,
        memory_usage,
    ]
    params_str = (
        "description, object_type, pu_name, subscription_type, "
        "subscription_recipient, memory_usage"
    )
    if any(unsupported_parameters):
        msg = (
            f"Passed unsupported parameter for this version of IServer "
            f"({ISERVER_VERSION_11_3_2}) Parameters supported only in 11.3.3+ "
            f"versions: {params_str}"
        )
        raise VersionException(msg)

    if elapsed_time:
        filters['duration'] = elapsed_time
    return filters


def __elapsed_filtering(elapsed: str) -> str:
    if (
        isinstance(elapsed, str)
        and len(elapsed) > 3
        and (elapsed.startswith('gt:') or elapsed.startswith('lt:'))
    ):
        if elapsed.startswith('gt'):
            return f'>{elapsed[3:]}'
        elif elapsed.startswith('lt'):
            return f'<{elapsed[3:]}'
        else:
            return elapsed
    else:
        raise TypeError('Incorrect "duration" format, correct example: "gt:100" ')


@class_version_handler('11.3.0200')
class Job(EntityBase):
    """Python representation of a Job object.
    Note: Some functionality is missing when working with 11.3.2 I-Server
        versions. To use all functionality 11.3.3 and up will be required.
        There are also some differences between 11.3.2 and 11.3.3+ Job
        attributes values.

    Attributes:
        connection: A Strategy One connection object
        id: Job information id
        description: Description of the job
        status: Status of the job
        type: Type of the job
        object_id: Id of object
        object_name: Name of object
        object_type: Type of object
        parent_id: Parent ID
        childs_ids: Array of children IDs
        subscription_recipient: Subscription recipient of the job
        subscription_type: Subscription delivery type of the job
        processing_unit_priority: Job Processing Unit Priority
        step_id: Step ID of the job
        user: Full name of the job initiator
        user_id: ID of the job initiator
        username: Username of the job initiator
        project_name: Name of the project in which the job is active
        project_id: Id of the project in which the job is active
        pu_name: Processing Unit name
        creation_time: Creation time of the job
        elapsed_time: Time elapsed since the creation of the job in milliseconds
            (11.3.3+) or seconds (11.3.2)
        step_elapsed_time: Time elapsed since the creation of the step
            in milliseconds
        memory_usage: Memory usage of the job in bytes
        client_machine: Client machine running the job
        total_tasks: Total tasks of the job
        completed_tasks: Completed tasks of the job
        error_message: Error message for the job
        error_time: Error time of the job
        filter_name: Name of the filter for the job
        sql: SQL of the job
        subscription_owner: Subscription owner
        template_name: Name of the template for the job
        step_statistics: Step statistics for the job
    """

    _FROM_DICT_MAP = {
        'creation_time': DatetimeFormats.FULLDATETIME,
        'error_time': DatetimeFormats.FULLDATETIME,
        'type': JobType,
        'status': JobStatus,
        'object_type': ObjectType,
        'subscription_type': SubscriptionType,
        'pu_name': PUName,
    }
    _API_GETTERS = {
        (
            'id',
            'type',
            'status',
            'user',
            'description',
            'object_type',
            'object_id',
            'object_name',
            'parent_id',
            'child_ids',
            'subscription_type',
            'processing_unit_priority',
            'step_id',
            'pu_name',
            'creation_time',
            'completed_tasks',
            'total_tasks',
            'project_name',
            'subscription_recipient',
            'memory_usage',
            'elapsed_time',
            'step_elapsed_time',
            'filter_name',
            'template_name',
            'sql',
            'subscription_owner',
            'error_time',
            'error_message',
            'step_statistics',
        ): monitors.get_job_v2
    }
    _REST_ATTR_MAP = {
        "job_id": None,  # delete job_id and only use id
        "job_type": "type",
        "user_full_name": "user",
        "total_task": "total_tasks",
        "completed_task": "completed_tasks",
        "filter": "filter_name",
        "template": "template_name",
        "machine": "client_machine",
        "priority": None,
        "warehouse_priority": None,
        "destination": None,
        "project_id": None,
        "username": None,
        "delivery_type": None,
        "duration": "elapsed_time",
    }

    def __init__(self, connection: Connection, id: str) -> None:
        """Initialize the Job object, populates it with I-Server data.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
            id: Job information id
        """
        Job._API_GETTERS = _set_api_wrappers(connection)
        super().__init__(connection, object_id=id)

    def _init_variables(self, **kwargs):
        """Initialize variables given kwargs.

        Note: attributes not accepted by any implementation of this function
            in the inheritance chain will be disregarded.
        """
        kwargs = self._rest_to_python(kwargs)
        # create _AVAILABLE_ATTRIBUTES map
        self._AVAILABLE_ATTRIBUTES.update(
            {key: type(val) for key, val in kwargs.items()}
        )

        self._connection = kwargs.get('connection')
        self._id = kwargs.get('id')
        self._description = kwargs.get('description')
        self._status = JobStatus(kwargs.get('status')) if kwargs.get('status') else None
        self._processing_unit_priority = kwargs.get('processing_unit_priority')
        self._creation_time = map_str_to_datetime(
            "creation_time", kwargs.get("creation_time"), self._FROM_DICT_MAP
        )
        self._elapsed_time = kwargs.get('elapsed_time')
        self._project_name = kwargs.get('project_name')
        self._object_id = kwargs.get('object_id')
        self._object_type = (
            ObjectType(kwargs.get('object_type')) if kwargs.get('object_type') else None
        )
        self._object_name = kwargs.get('object_name')
        self._sql = kwargs.get('sql')
        self._subscription_owner = kwargs.get('subscription_owner')
        self._subscription_recipient = kwargs.get('subscription_recipient')
        self._client_machine = kwargs.get('client_machine')
        self._type = JobType(kwargs.get('type')) if kwargs.get('type') else None
        self._user = kwargs.get('user')
        self._total_tasks = kwargs.get('total_tasks')
        self._completed_tasks = kwargs.get('completed_tasks')
        self._parent_id = kwargs.get('parent_id')
        self._child_ids = kwargs.get('child_ids')
        self._subscription_type = (
            SubscriptionType(kwargs.get('subscription_type'))
            if kwargs.get('subscription_type')
            else kwargs.get('subscription_type')
        )
        self._step_id = kwargs.get('step_id')
        self._pu_name = PUName(kwargs.get('pu_name')) if kwargs.get('pu_name') else None
        self._memory_usage = kwargs.get('memory_usage')
        self._step_elapsed_time = kwargs.get('step_elapsed_time')
        self._filter_name = kwargs.get('filter_name')
        self._template_name = kwargs.get('template_name')
        self._error_time = (
            map_str_to_datetime(
                "error_time", kwargs.get("error_time"), self._FROM_DICT_MAP
            )
            if kwargs.get("error_time")
            else None
        )
        self._error_message = kwargs.get('error_message')
        self._step_statistics = kwargs.get('step_statistics')
        self._user_id = None
        self._username = None

    def kill(self) -> bool:
        """Kill the job.

        Returns:
            True if successfully killed job, False otherwise.
        """
        response = monitors.cancel_job(self._connection, str(self.id))
        if config.verbose:
            success_msg = f"Successfully killed {self.id}."
            fail_msg = f"Error killing job {self.id}."
            msg = success_msg if response.ok else fail_msg
            logger.info(msg)
        return response.ok

    def list_properties(self, excluded_properties: list[str] | None = None) -> dict:
        """List Job object properties."""
        properties = super().list_properties(excluded_properties)
        properties.pop('connection', None)
        return properties

    def refresh_status(self) -> None:
        """Refresh the job status."""
        try:
            self.fetch()
        except IServerError as e:
            if e.http_code == 500 or e.http_code == 404:
                stable_states = [
                    JobStatus.COMPLETED,
                    JobStatus.ERROR,
                    JobStatus.STOPPED,
                ]
                if self._status and self._status not in stable_states:
                    self._status = JobStatus.COMPLETED

    @property
    def status(self):
        return self._status

    @property
    def description(self):
        return self._description

    @property
    def processing_unit_priority(self):
        return self._processing_unit_priority

    @property
    def creation_time(self):
        return self._creation_time

    @property
    def project_name(self):
        return self._project_name

    @property
    def object_type(self):
        return self._object_type

    @property
    def object_id(self):
        return self._object_id

    @property
    def sql(self):
        return self._sql

    @property
    def subscription_owner(self):
        return self._subscription_owner

    @property
    def subscription_recipient(self):
        return self._subscription_recipient

    @property
    def client_machine(self):
        return self._client_machine

    @property
    def user(self):
        return self._user

    @property
    def total_tasks(self):
        return self._total_tasks

    @property
    def completed_tasks(self):
        return self._completed_tasks

    @property
    def parent_id(self):
        return self._parent_id

    @property
    def child_ids(self):
        return self._child_ids

    @property
    def subscription_type(self):
        return self._subscription_type

    @property
    def step_id(self):
        return self._step_id

    @property
    def pu_name(self):
        return self._pu_name

    @property
    def memory_usage(self):
        return self._memory_usage

    @property
    def elapsed_time(self):
        return self._elapsed_time

    @property
    def step_elapsed_time(self):
        return self._step_elapsed_time

    @property
    def filter_name(self):
        return self._filter_name

    @property
    def template_name(self):
        return self._template_name

    @property
    def error_time(self):
        return self._error_time

    @property
    def error_message(self):
        return self._error_message

    @property
    def step_statistics(self):
        return self._step_statistics

    @property
    def object_name(self):
        return self._object_name

    @property
    def user_id(self):
        from mstrio.users_and_groups import User

        if self._user:
            if not self._user_id:
                users = User._get_user_ids(
                    connection=self._connection,
                    name_begins=self._user,
                    name=self._user,
                )
                if users:
                    [self._user_id] = users
                else:
                    exception_handler(
                        f"There is no user: '{self._user}'", exception_type=ValueError
                    )
            return self._user_id
        else:
            return None

    @property
    def username(self):
        from mstrio.users_and_groups import User

        if self._user:
            if not self._username:
                if self._user_id:
                    self._username = User(self._connection, id=self._user_id).username
                else:
                    user = User(self._connection, name=self._user)
                    self._username = user.username
                    self._user_id = user.id
            return self._username
        else:
            return None



"""This is the demo script to show how administrator can monitor and manage jobs

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.

Note: Some functionality is missing when working with 11.3.2 I-Server
    versions. To use all functionality 11.3.3 and up will be required.
    There are also some differences between 11.3.2 and 11.3.3+ Job
    attributes values.
"""

from mstrio.server import (
    Job, JobStatus, JobType, kill_all_jobs, kill_jobs, list_jobs, ObjectType, PUName
)
from mstrio.server import Project
from mstrio.users_and_groups import User

from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# get project by name
project = Project(connection=conn, name=PROJECT_NAME)

# get list of `Job` objects for all jobs on the environment
jobs = list_jobs(conn)
print(jobs)

# get list of dicts representing job information for all jobs on the environment
jobs = list_jobs(conn, to_dictionary=True)
print(jobs)

# Define variables which can be later used in a script
JOB_ID = $job_id  # id for Job object lookup

# instantiate an existing job using constructor
job = Job(conn, id=JOB_ID)

# get properties for the first job on the list
job.list_properties()

# kill the job
job.kill()

# get a list of `Job` objects filtered by job status, project and object type
list_jobs(conn, status=JobStatus.LOADING_PROMPT, project=project, object_type=ObjectType.CUBE)
# see server/job_monitor.py for JobStatus and ObjectType values

# Define variables which can be later used in a script
USERNAME = $username  # name for User object lookup

# get user by name
john = User(connection=conn, name=USERNAME)

# get a list of `Job` objects filtered by job type and job owner
list_jobs(
    conn, type=JobType.INTERACTIVE, user=john
)  # see server/job_monitor.py for JobType values

# Define variables which can be later used in a script
LIST_JOBS_FILTER_TIME = $list_jobs_filter_time # for example 'gt:10000'
LIST_JOBS_FILTER_MEM = $list_jobs_filter_mem # for example 'gt:500'

# get a list of `Job` objects filtered by elapsed time and memory usage
# NOTE: memory_usage filter works for 11.3.3+ I-Server versions
slow_jobs = list_jobs(conn, elapsed_time=LIST_JOBS_FILTER_TIME, memory_usage=LIST_JOBS_FILTER_MEM)

# kill jobs by passing either Job objects or ids
result = kill_jobs(conn, slow_jobs)

# kill_jobs return Success, PartialSuccess or MSTRException, you can check which
# jobs were killed and which were not, and why in case of PartialSuccess
print(result.succeeded)
print(result.failed)

# you can easily evaluate if all passed jobs were killed using bool() evaluation
# on result, which will return True if result is Success and False if result
# is PartialSuccess
successfully_killed_passed_jobs = bool(result)
if result:
    print('Successfully killed jobs.')
    # perform other automation actions

# Define a variable which can be later used in a script
KILL_JOBS_FILTER_TIME = $kill_jobs_filter_time  # should be converted to seconds

# kill jobs running over a certain time
elapsed_t = KILL_JOBS_FILTER_TIME  # time in seconds (for 11.3.2 I-Server version)
if not conn.iserver_version or conn.iserver_version >= '11.3.0300':
    elapsed_t = 1000 * elapsed_t  # convert to milliseconds (for 11.3.3+ I-Server versions)
elapsed_t = f'gt:{elapsed_t}'  # correct form filter (valid operators: 'gt:' and 'lt:')
kill_all_jobs(conn, elapsed_time=elapsed_t, force=True)

# kill filtered jobs similarly to list_jobs() using single function
result = kill_all_jobs(conn, pu_name=PUName.SQL_ENGINE, memory_usage='gt:800')
# see server/job_monitor.py for PUName values

# like kill_jobs, kill_all_jobs return Success, PartialSuccess or MSTRException
print(result.succeeded)



"""Remove each address from every departed user (those which are disabled).
For each successfully removed address a message will be printed.

1. Connect to the environment using data from workstation
2. Get all users that are disabled
3. For every retrieved user remove all its email addresses
4. Return list which contains information about removed addresses such as user's id,
   username and address
"""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import list_users


def delete_addresses_from_departed_users(connection: "Connection") -> List[dict]:
    """Remove each address from every departed user (those which are disabled).
    For each successfully removed address a message will be printed.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
    """

    # get all users that are disabled
    all_users = list_users(connection=connection)
    users_ = [u for u in all_users if not u.enabled]
    removed_addresses = []
    for user_ in users_:
        # remove all email addresses from the given user
        if user_.addresses:
            for address in user_.addresses:
                user_.remove_address(id=address['id'])

                removed_addresses.append(
                    {
                        'user_id': user_.id,
                        'username': user_.username,
                        'address': address,
                    }
                )
    return removed_addresses


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# execute deletion of addresses from departed users
delete_addresses_from_departed_users(conn)



# Privileges
from mstrio.access_and_security.privilege import Privilege
from mstrio.users_and_groups import list_user_groups, list_users, User, UserGroup
from mstrio.access_and_security.security_role import list_security_roles, SecurityRole
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
PRIVILEGE_NAME = $privilege_name  # Insert name of edited privilege here
PRIVILEGE_ID = $privilege_id  # Insert ID of edited privilege here

# get Privilege object by name
priv = Privilege(conn, name=PRIVILEGE_NAME)
# get Privilege object by id
priv = Privilege(conn, id=PRIVILEGE_ID)

# List Privileges and return objects or display in DataFrame
Privilege.list_privileges(conn, to_dataframe=True, is_project_level_privilege='True')
priv = Privilege.list_privileges(conn, id=[PRIVILEGE_ID])
for p in priv:
    print(p.id)
priv[0].list_properties()

# Define variables which can be later used in a script
SECURITY_ROLE_NAME = $security_role_name  # Insert name of newly created or accessed security role
SECURITY_ROLE_DESCRIPTION = $security_role_description  # Insert description of newly created or accessed security role

# SecurityRoles
# Create new SecurityRole
new_role = SecurityRole.create(
    conn,
    name=SECURITY_ROLE_NAME,
    description=SECURITY_ROLE_DESCRIPTION,
    privileges=[PRIVILEGE_ID, PRIVILEGE_NAME]
)

# List SecurityRoles and store the Objects
all_roles = list_security_roles(conn)
list_security_roles(conn, to_dataframe=True)

# Create SecurityRole object by name or ID
role = SecurityRole(conn, id=all_roles[0].id)
SecurityRole(conn, id=all_roles[0].id)
role = SecurityRole(conn, name=all_roles[0].name)
SecurityRole(conn, name=all_roles[0].name)
SecurityRole(connection=conn, name=SECURITY_ROLE_NAME)

# List SecurityRole members
role.list_members(project_name=PROJECT_NAME)

# Define variables which can be later used in a script
USERNAME = $username  # Insert name of user to be assigned or revoked security role
USER_GROUP_NAME = $user_group_name  # Insert name of user group to be assigned or revoked security role

# Grant/Revoke Security Role to users/usergroups
user = User(conn, name=USERNAME)
users = list_users(conn)
group = UserGroup(conn, name=USER_GROUP_NAME)
groups = list_user_groups(conn)

# Grant/Revoke for Users
role.grant_to(members=user, project=PROJECT_NAME)
role.revoke_from(members=user, project=PROJECT_NAME)
role.grant_to(members=users, project=PROJECT_NAME)
role.revoke_from(members=users, project=PROJECT_NAME)

# Grant/Revoke for UserGroups
role.grant_to(members=group, project=PROJECT_NAME)
role.revoke_from(members=group, project=PROJECT_NAME)

# List Privileges
role.list_privileges()
role.list_privileges(to_dataframe=True)

# Grant/ Revoke privileges to Security Role
role.grant_privilege(privilege=[PRIVILEGE_ID])
role.revoke_privilege([PRIVILEGE_ID])
privs = list(role.list_privileges().keys())
role.revoke_all_privileges()
role.grant_privilege(privilege=privs)



"""
Manage Pre-/Post-execution SQL statements with Subscription execution.
Check the ETL flag in the database, run Subscription,
and send an email to the Administrator.

1. Connect to environment using data from workstation
2. Get datasource instance base on provided name ('Tutorial Postgres')
3. Get project object based on provided name ('MicroStrategy Tutorial')
4. Steps below describe the process of managing email subscription for the given
   project and datasource instance
5. Prepare query to select the ETL flag 
6. Execute the query and store the result
7. If the ETL flag is present, get the email subscription object by an id
   and execute it
8. If the ETL flag is missing, send an email to the administrator
9. Steps below describe the process of managing mobile subscription for the given
   project and datasource instance
10. Get the mobile subscription object by an id
11. Execute the mobile subscription and wait for its completion checking the status
12. Prepare SQL query to insert status of the mobile subscription into the table
13. Execute the query
14. Calculate the success rate of the mobile subscription
15. Prepare email content with the results of the mobile subscription's execution
    and send it to the administrator
"""

import time

from mstrio.connection import Connection, get_connection
from mstrio.datasources import DatasourceInstance
from mstrio.distribution_services import EmailSubscription, MobileSubscription
from mstrio.distribution_services.email import send_email
from mstrio.server import Project
from mstrio.users_and_groups import User


def send_email_to_administrator(connection: Connection, subject: str, content: str):
    """Send an email to the Administrator.

    Args:
        connection: Strategy One connection object.
        subject: Email subject.
        content: Email content.
    """
    administrator = User(connection=connection, name="Administrator")
    send_email(
        connection=connection,
        subject=subject,
        content=content,
        users=[administrator],
        is_html=True,
    )


def manage_email_subscription(
    connection: 'Connection',
    datasource_instance: 'DatasourceInstance',
    project: 'Project',
):
    """Manage the Email Subscription.

    Args:
        connection: Strategy One connection object.
        datasource_instance: Datasource instance object.
        project: Project object.
    """
    # SQL query to select the ETL flag
    select_query = 'SELECT <ETL_FLAG> FROM <TABLE_NAME>'
    # Execute the SQL query
    sql_result = datasource_instance.execute_query(
        project_id=project.id,
        query=select_query,
        max_retries=10,
        retry_delay=10,
    )
    # If the ETL flag is present, run the Email Subscription
    if sql_result and sql_result.get('results'):
        print('ETL flag is present. Running the Email Subscription.')
        email_subscription = EmailSubscription(
            connection=conn, id="7138EE3011EA0FD26B470080EF752E56"
        )
        email_subscription.execute()
    # If the ETL flag is missing, send an email to the Administrator
    else:
        print("The ETL flag is missing. Sending an email to the Administrator.")
        send_email_to_administrator(
            connection=connection,
            subject="ETL Flag Missing",
            content="The ETL flag is missing. Please check the ETL process.",
        )


def manage_mobile_subscription(
    connection: 'Connection',
    datasource_instance: 'DatasourceInstance',
    project: 'Project',
):
    """Manage the Mobile Subscription.

    Args:
        connection: Strategy One connection object.
        datasource_instance: Datasource instance object.
        project: Project object.
    """
    # Table name to save the Mobile Subscription status
    TABLE_NAME = "<TABLE_NAME>"

    mobile_subscription = MobileSubscription(
        connection=connection, id="EBC6F6D7F3413F546285B091B746FC69"
    )
    mobile_subscription.execute()

    # Wait for the Mobile Subscription to complete
    for _ in range(10):
        if mobile_subscription.status is None:
            print("Mobile Subscription is still running. Waiting for completion.")
            time.sleep(10)
        break

    mobile_subscription_status = mobile_subscription.status.to_dict(False).values()

    # Convert each item in the list to a string
    # and replace single quotes with double quotes
    str_mobile_subscription_status = [
        str(x).replace("'", '"') for x in mobile_subscription_status
    ]

    # SQL query to insert the Mobile Subscription status into the table
    insert_query = f"""
    INSERT INTO {TABLE_NAME}
    VALUES
        ({", ".join([f"'{x}'" for x in str_mobile_subscription_status])});
    """

    # Execute the SQL query
    datasource_instance.execute_query(
        project_id=project.id,
        query=insert_query,
    )

    # Calculate the success rate of the Mobile Subscription
    result = sum(
        1 for status in mobile_subscription.status.statuses if status.error is None
    )
    success_rate = (result / mobile_subscription.status.total) * 100 if result else 0

    email_content = f"""
    <html>
        <body>
            <h1>Subscription with name: {mobile_subscription.name} was executed</h1>
            <p>Total number of successful deliveries: {result}</p>
            <p>Total number of deliveries: {mobile_subscription.status.total}</p>
            <p>Successful rate: {success_rate}% </p>
            <p>All details were saved to the table {TABLE_NAME}</p>
        </body>
    </html>
    """

    send_email_to_administrator(
        connection=connection, subject="Mobile Subscription", content=email_content
    )


def manage_execution(connection: 'Connection'):
    """Manage Pre-/Post-execution SQL statements with Subscription execution.

    Args:
        connection: Strategy One connection object.
    """
    datasource_instance = DatasourceInstance(
        connection=connection, name='Tutorial Postgres'
    )
    project = Project(connection=connection, name='MicroStrategy Tutorial')
    manage_email_subscription(connection, datasource_instance, project)
    manage_mobile_subscription(connection, datasource_instance, project)


# Create connection based on workstation data
conn = get_connection(workstationData, 'MicroStrategy Tutorial')
manage_execution(conn)
